# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit tests ONTAP Ansible module: na_ontap_snapshot_policy'''

from __future__ import print_function
import json
import pytest

from units.compat import unittest
from units.compat.mock import patch, Mock
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_snapshot_policy \
    import NetAppOntapSnapshotPolicy as my_module

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None):
        ''' save arguments '''
        self.type = kind
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.type == 'policy':
            xml = self.build_snapshot_policy_info()
        elif self.type == 'policy_fail':
            raise netapp_utils.zapi.NaApiError(code='TEST', message="This exception is from the unit test")
        self.xml_out = xml
        return xml

    def asup_log_for_cserver(self):
        ''' mock autosupport log'''
        return None

    @staticmethod
    def build_snapshot_policy_info():
        ''' build xml data for snapshot-policy-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        data = {'num-records': 1,
                'attributes-list': {'snapshot-policy-info': {'policy': 'ansible'}}}
        xml.translate_struct(data)
        return xml


class TestMyModule(unittest.TestCase):
    ''' a group of related Unit Tests '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.server = MockONTAPConnection()
        self.onbox = False

    def set_default_args(self):
        if self.onbox:
            hostname = '10.10.10.10'
            username = 'admin'
            password = '1234'
            name = 'ansible'
            enabled = True
            count = 100
            schedule = 'hourly'
            comment = 'new comment'
        else:
            hostname = 'hostname'
            username = 'username'
            password = 'password'
            name = 'ansible'
            enabled = True
            count = 100
            schedule = 'hourly'
            comment = 'new comment'
        return dict({
            'hostname': hostname,
            'username': username,
            'password': password,
            'name': name,
            'enabled': enabled,
            'count': count,
            'schedule': schedule,
            'comment': comment
        })

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            my_module()
        print('Info: %s' % exc.value.args[0]['msg'])

    def test_ensure_get_called(self):
        ''' test get_snapshot_policy()  for non-existent snapshot policy'''
        set_module_args(self.set_default_args())
        my_obj = my_module()
        my_obj.server = self.server
        assert my_obj.get_snapshot_policy() is None

    def test_ensure_get_called_existing(self):
        ''' test get_snapshot_policy()  for existing snapshot policy'''
        set_module_args(self.set_default_args())
        my_obj = my_module()
        my_obj.server = MockONTAPConnection(kind='policy')
        assert my_obj.get_snapshot_policy()

    @patch('ansible.modules.storage.netapp.na_ontap_snapshot_policy.NetAppOntapSnapshotPolicy.create_snapshot_policy')
    def test_successful_create(self, create_snapshot):
        ''' creating snapshot policy and testing idempotency '''
        set_module_args(self.set_default_args())
        my_obj = my_module()
        my_obj.asup_log_for_cserver = Mock(return_value=None)
        if not self.onbox:
            my_obj.server = self.server
        with pytest.raises(AnsibleExitJson) as exc:
            my_obj.apply()
        assert exc.value.args[0]['changed']
        create_snapshot.assert_called_with()
        # to reset na_helper from remembering the previous 'changed' value
        my_obj = my_module()
        my_obj.asup_log_for_cserver = Mock(return_value=None)
        if not self.onbox:
            my_obj.server = MockONTAPConnection('policy')
        with pytest.raises(AnsibleExitJson) as exc:
            my_obj.apply()
        assert not exc.value.args[0]['changed']

    @patch('ansible.modules.storage.netapp.na_ontap_snapshot_policy.NetAppOntapSnapshotPolicy.delete_snapshot_policy')
    def test_successful_delete(self, delete_snapshot):
        ''' deleting snapshot policy and testing idempotency '''
        data = self.set_default_args()
        data['state'] = 'absent'
        set_module_args(data)
        my_obj = my_module()
        my_obj.asup_log_for_cserver = Mock(return_value=None)
        if not self.onbox:
            my_obj.server = MockONTAPConnection('policy')
        with pytest.raises(AnsibleExitJson) as exc:
            my_obj.apply()
        assert exc.value.args[0]['changed']
        delete_snapshot.assert_called_with()
        # to reset na_helper from remembering the previous 'changed' value
        my_obj = my_module()
        my_obj.asup_log_for_cserver = Mock(return_value=None)
        if not self.onbox:
            my_obj.server = self.server
        with pytest.raises(AnsibleExitJson) as exc:
            my_obj.apply()
        assert not exc.value.args[0]['changed']

    def test_if_all_methods_catch_exception(self):
        module_args = {}
        module_args.update(self.set_default_args())
        set_module_args(module_args)
        my_obj = my_module()
        if not self.onbox:
            my_obj.server = MockONTAPConnection('policy_fail')
        with pytest.raises(AnsibleFailJson) as exc:
            my_obj.create_snapshot_policy()
        assert 'Error creating snapshot policy ansible:' in exc.value.args[0]['msg']
        with pytest.raises(AnsibleFailJson) as exc:
            my_obj.delete_snapshot_policy()
        assert 'Error deleting snapshot policy ansible:' in exc.value.args[0]['msg']



# (c) 2019, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit test template for ONTAP Ansible module '''

from __future__ import print_function
import json
import pytest

from units.compat import unittest
from units.compat.mock import patch, Mock
from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes
import ansible.module_utils.netapp as netapp_utils

from ansible.modules.storage.netapp.na_ontap_security_key_manager \
    import NetAppOntapSecurityKeyManager as key_manager_module  # module under test

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip('skipping as missing required netapp_lib')


def set_module_args(args):
    """prepare arguments so that they will be picked up during module creation"""
    args = json.dumps({'ANSIBLE_MODULE_ARGS': args})
    basic._ANSIBLE_ARGS = to_bytes(args)  # pylint: disable=protected-access


class AnsibleExitJson(Exception):
    """Exception class to be raised by module.exit_json and caught by the test case"""
    pass


class AnsibleFailJson(Exception):
    """Exception class to be raised by module.fail_json and caught by the test case"""
    pass


def exit_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over exit_json; package return data into an exception"""
    if 'changed' not in kwargs:
        kwargs['changed'] = False
    raise AnsibleExitJson(kwargs)


def fail_json(*args, **kwargs):  # pylint: disable=unused-argument
    """function to patch over fail_json; package return data into an exception"""
    kwargs['failed'] = True
    raise AnsibleFailJson(kwargs)


class MockONTAPConnection(object):
    ''' mock server connection to ONTAP host '''

    def __init__(self, kind=None, data=None):
        ''' save arguments '''
        self.type = kind
        self.data = data
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.type == 'key_manager':
            xml = self.build_port_info(self.data)
        self.xml_out = xml
        return xml

    @staticmethod
    def build_port_info(key_manager_details):
        ''' build xml data for-key-manager-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = {
            'num-records': 1,
            'attributes-list': {
                'key-manager-info': {
                    'key-manager-ip-address': '0.0.0.0',
                    'key-manager-server-status': 'available',
                    'key-manager-tcp-port': '5696',
                    'node-name': 'test_node'
                }
            }
        }
        xml.translate_struct(attributes)
        return xml


class TestMyModule(unittest.TestCase):
    ''' a group of related Unit Tests '''

    def setUp(self):
        self.mock_module_helper = patch.multiple(basic.AnsibleModule,
                                                 exit_json=exit_json,
                                                 fail_json=fail_json)
        self.mock_module_helper.start()
        self.addCleanup(self.mock_module_helper.stop)
        self.mock_key_manager = {
            'node_name': 'test_node',
            'tcp_port': 5696,
            'ip_address': '0.0.0.0',
            'server_status': 'available'
        }

    def mock_args(self):
        return {
            'node': self.mock_key_manager['node_name'],
            'tcp_port': self.mock_key_manager['tcp_port'],
            'ip_address': self.mock_key_manager['ip_address'],
            'hostname': 'test',
            'username': 'test_user',
            'password': 'test_pass!',
            'https': 'False'
        }

    def get_key_manager_mock_object(self, kind=None):
        """
        Helper method to return an na_ontap_security_key_manager object
        :param kind: passes this param to MockONTAPConnection()
        :return: na_ontap_security_key_manager object
        """
        obj = key_manager_module()
        obj.asup_log_for_cserver = Mock(return_value=None)
        obj.cluster = Mock()
        obj.cluster.invoke_successfully = Mock()
        if kind is None:
            obj.cluster = MockONTAPConnection()
        else:
            obj.cluster = MockONTAPConnection(kind=kind, data=self.mock_key_manager)
        return obj

    def test_module_fail_when_required_args_missing(self):
        ''' required arguments are reported as errors '''
        with pytest.raises(AnsibleFailJson) as exc:
            set_module_args({})
            key_manager_module()
        print('Info: %s' % exc.value.args[0]['msg'])

    def test_get_nonexistent_key_manager(self):
        ''' Test if get_key_manager() returns None for non-existent key manager '''
        set_module_args(self.mock_args())
        result = self.get_key_manager_mock_object().get_key_manager()
        assert result is None

    def test_get_existing_key_manager(self):
        ''' Test if get_key_manager() returns details for existing key manager '''
        set_module_args(self.mock_args())
        result = self.get_key_manager_mock_object('key_manager').get_key_manager()
        assert result['ip_address'] == self.mock_key_manager['ip_address']

    @patch('ansible.modules.storage.netapp.na_ontap_security_key_manager.NetAppOntapSecurityKeyManager.get_key_manager')
    def test_successfully_add_key_manager(self, get_key_manager):
        ''' Test successfully add key manager'''
        data = self.mock_args()
        data['state'] = 'present'
        set_module_args(data)
        get_key_manager.side_effect = [
            None
        ]
        obj = self.get_key_manager_mock_object('key_manager')
        with pytest.raises(AnsibleExitJson) as exc:
            obj.apply()
        assert exc.value.args[0]['changed']

    def test_successfully_delete_key_manager(self):
        ''' Test successfully delete key manager'''
        data = self.mock_args()
        data['state'] = 'absent'
        set_module_args(data)
        obj = self.get_key_manager_mock_object('key_manager')
        with pytest.raises(AnsibleExitJson) as exc:
            obj.apply()
        assert exc.value.args[0]['changed']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


try:
    from library.module_utils.network.f5.common import F5BaseClient
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.icontrol import iControlRestSession
except ImportError:
    from ansible.module_utils.network.f5.common import F5BaseClient
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.icontrol import iControlRestSession


class F5RestClient(F5BaseClient):
    def __init__(self, *args, **kwargs):
        super(F5RestClient, self).__init__(*args, **kwargs)
        self.provider = self.merge_provider_params()
        self.headers = {
            'Content-Type': 'application/json'
        }

    @property
    def api(self):
        if self._client:
            return self._client
        session, err = self.connect_via_token_auth()
        if err or session is None:
            session, err = self.connect_via_basic_auth()
            if err or session is None:
                raise F5ModuleError(err)
        self._client = session
        return session

    def connect_via_token_auth(self):
        url = "https://{0}:{1}/mgmt/shared/authn/login".format(
            self.provider['server'], self.provider['server_port']
        )
        payload = {
            'username': self.provider['user'],
            'password': self.provider['password'],
            'loginProviderName': self.provider['auth_provider'] or 'tmos'
        }
        session = iControlRestSession(
            validate_certs=self.provider['validate_certs']
        )

        response = session.post(
            url,
            json=payload,
            headers=self.headers
        )

        if response.status not in [200]:
            return None, response.content

        session.request.headers['X-F5-Auth-Token'] = response.json()['token']['token']
        return session, None

    def connect_via_basic_auth(self):
        url = "https://{0}:{1}/mgmt/tm/sys".format(
            self.provider['server'], self.provider['server_port']
        )
        session = iControlRestSession(
            url_username=self.provider['user'],
            url_password=self.provider['password'],
            validate_certs=self.provider['validate_certs'],
        )

        response = session.get(
            url,
            headers=self.headers
        )

        if response.status not in [200]:
            return None, response.content
        return session, None



#!/usr/bin/python
# Copyright: (c) 2018, Pluribus Networks
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: pn_admin_service
author: "Pluribus Networks (@rajaspachipulusu17)"
version_added: "2.8"
short_description: CLI command to modify admin-service
description:
  - This module is used to modify services on the server-switch.
options:
  pn_cliswitch:
    description:
      - Target switch to run the CLI on.
    required: False
    type: str
  state:
    description:
      - State the action to perform. Use C(update) to modify the admin-service.
    required: True
    type: str
    choices: ['update']
  pn_web:
    description:
      - Web (HTTP) to enable or disable.
    required: False
    type: bool
  pn_web_ssl:
    description:
      - Web SSL (HTTPS) to enable or disable.
    required: False
    type: bool
  pn_snmp:
    description:
      - Simple Network Monitoring Protocol (SNMP) to enable or disable.
    required: False
    type: bool
  pn_web_port:
    description:
      - Web (HTTP) port to enable or disable.
    required: False
    type: str
  pn_web_ssl_port:
    description:
      - Web SSL (HTTPS) port to enable or disable.
    required: False
    type: str
  pn_nfs:
    description:
      - Network File System (NFS) to enable or disable.
    required: False
    type: bool
  pn_ssh:
    description:
      - Secure Shell to enable or disable.
    required: False
    type: bool
  pn_web_log:
    description:
      - Web logging to enable or disable.
    required: False
    type: bool
  pn__if:
    description:
      - administrative service interface.
    required: False
    type: str
    choices: ['mgmt', 'data']
  pn_icmp:
    description:
      - Internet Message Control Protocol (ICMP) to enable or disable.
    required: False
    type: bool
  pn_net_api:
    description:
      - Netvisor API to enable or disable APIs.
    required: False
    type: bool
"""

EXAMPLES = """
- name: admin service functionality
  pn_admin_service:
    pn_cliswitch: "sw01"
    state: "update"
    pn__if: "mgmt"
    pn_web: False
    pn_icmp: True

- name: admin service functionality
  pn_admin_service:
    pn_cliswitch: "sw01"
    state: "update"
    pn_web: False
    pn__if: "mgmt"
    pn_snmp: True
    pn_net_api: True
    pn_ssh: True
"""

RETURN = """
command:
  description: the CLI command run on the target node.
  returned: always
  type: str
stdout:
  description: set of responses from the admin-service command.
  returned: always
  type: list
stderr:
  description: set of error responses from the admin-service command.
  returned: on error
  type: list
changed:
  description: indicates whether the CLI caused changes on the target.
  returned: always
  type: bool
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.netvisor.pn_nvos import pn_cli, booleanArgs, run_cli


def main():
    """ This section is for arguments parsing """

    state_map = dict(
        update='admin-service-modify'
    )

    module = AnsibleModule(
        argument_spec=dict(
            pn_cliswitch=dict(required=False, type='str'),
            state=dict(required=True, type='str',
                       choices=state_map.keys()),
            pn_web=dict(required=False, type='bool'),
            pn_web_ssl=dict(required=False, type='bool'),
            pn_snmp=dict(required=False, type='bool'),
            pn_web_port=dict(required=False, type='str'),
            pn_web_ssl_port=dict(required=False, type='str'),
            pn_nfs=dict(required=False, type='bool'),
            pn_ssh=dict(required=False, type='bool'),
            pn_web_log=dict(required=False, type='bool'),
            pn__if=dict(required=False, type='str', choices=['mgmt', 'data']),
            pn_icmp=dict(required=False, type='bool'),
            pn_net_api=dict(required=False, type='bool'),
        ),
        required_if=([['state', 'update', ['pn__if']]]),
        required_one_of=[['pn_web', 'pn_web_ssl', 'pn_snmp',
                          'pn_web_port', 'pn_web_ssl_port', 'pn_nfs',
                          'pn_ssh', 'pn_web_log', 'pn_icmp', 'pn_net_api']]
    )

    # Accessing the arguments
    cliswitch = module.params['pn_cliswitch']
    state = module.params['state']
    web = module.params['pn_web']
    web_ssl = module.params['pn_web_ssl']
    snmp = module.params['pn_snmp']
    web_port = module.params['pn_web_port']
    web_ssl_port = module.params['pn_web_ssl_port']
    nfs = module.params['pn_nfs']
    ssh = module.params['pn_ssh']
    web_log = module.params['pn_web_log']
    _if = module.params['pn__if']
    icmp = module.params['pn_icmp']
    net_api = module.params['pn_net_api']

    command = state_map[state]

    # Building the CLI command string
    cli = pn_cli(module, cliswitch)

    if command == 'admin-service-modify':
        cli += ' %s ' % command

        if _if:
            cli += ' if ' + _if
        if web_port:
            cli += ' web-port ' + web_port
        if web_ssl_port:
            cli += ' web-ssl-port ' + web_ssl_port

        cli += booleanArgs(web, 'web', 'no-web')
        cli += booleanArgs(web_ssl, 'web-ssl', 'no-web-ssl')
        cli += booleanArgs(snmp, 'snmp', 'no-snmp')
        cli += booleanArgs(nfs, 'nfs', 'no-nfs')
        cli += booleanArgs(ssh, 'ssh', 'no-ssh')
        cli += booleanArgs(icmp, 'icmp', 'no-icmp')
        cli += booleanArgs(net_api, 'net-api', 'no-net-api')
        cli += booleanArgs(web_log, 'web-log', 'no-web-log')

    run_cli(module, cli, state_map)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright: (c) 2018, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import json
import pytest
import sys

if sys.version_info < (2, 7):
    pytestmark = pytest.mark.skip("F5 Ansible modules require Python >= 2.7")

from ansible.module_utils.basic import AnsibleModule

try:
    from library.modules.bigip_profile_fastl4 import ApiParameters
    from library.modules.bigip_profile_fastl4 import ModuleParameters
    from library.modules.bigip_profile_fastl4 import ModuleManager
    from library.modules.bigip_profile_fastl4 import ArgumentSpec

    # In Ansible 2.8, Ansible changed import paths.
    from test.units.compat import unittest
    from test.units.compat.mock import Mock
    from test.units.compat.mock import patch

    from test.units.modules.utils import set_module_args
except ImportError:
    from ansible.modules.network.f5.bigip_profile_fastl4 import ApiParameters
    from ansible.modules.network.f5.bigip_profile_fastl4 import ModuleParameters
    from ansible.modules.network.f5.bigip_profile_fastl4 import ModuleManager
    from ansible.modules.network.f5.bigip_profile_fastl4 import ArgumentSpec

    # Ansible 2.8 imports
    from units.compat import unittest
    from units.compat.mock import Mock
    from units.compat.mock import patch

    from units.modules.utils import set_module_args


fixture_path = os.path.join(os.path.dirname(__file__), 'fixtures')
fixture_data = {}


def load_fixture(name):
    path = os.path.join(fixture_path, name)

    if path in fixture_data:
        return fixture_data[path]

    with open(path) as f:
        data = f.read()

    try:
        data = json.loads(data)
    except Exception:
        pass

    fixture_data[path] = data
    return data


class TestParameters(unittest.TestCase):
    def test_module_parameters(self):
        args = dict(
            name='foo',
            parent='bar',
            idle_timeout=100,
            client_timeout=101,
            description='description one',
            explicit_flow_migration=False,
            ip_df_mode='pmtu',
            ip_tos_to_client=102,
            ip_tos_to_server=103,
            ip_ttl_v4=104,
            ip_ttl_v6=105,
            ip_ttl_mode='proxy',
            keep_alive_interval=106,
            late_binding=True,
            link_qos_to_client=7,
            link_qos_to_server=6,
            loose_close=False,
            loose_initialization=True,
            mss_override=4,
            reassemble_fragments=True,
            receive_window_size=109,
            reset_on_timeout=False,
            rtt_from_client=True,
            rtt_from_server=False,
            server_sack=True,
            server_timestamp=False,
            syn_cookie_mss=110,
            tcp_close_timeout=111,
            tcp_generate_isn=True,
            tcp_handshake_timeout=112,
            tcp_strip_sack=False,
            tcp_time_wait_timeout=113,
            tcp_timestamp_mode='rewrite',
            tcp_wscale_mode='strip',
            timeout_recovery='fallback',
        )

        p = ModuleParameters(params=args)
        assert p.name == 'foo'
        assert p.parent == '/Common/bar'
        assert p.description == 'description one'
        assert p.idle_timeout == 100
        assert p.client_timeout == 101
        assert p.explicit_flow_migration == 'no'
        assert p.ip_df_mode == 'pmtu'
        assert p.ip_tos_to_client == 102
        assert p.ip_tos_to_server == 103
        assert p.ip_ttl_v4 == 104
        assert p.ip_ttl_v6 == 105
        assert p.ip_ttl_mode == 'proxy'
        assert p.keep_alive_interval == 106
        assert p.late_binding == 'yes'
        assert p.link_qos_to_client == 7
        assert p.link_qos_to_server == 6
        assert p.loose_close == 'no'
        assert p.loose_initialization == 'yes'
        assert p.mss_override == 4
        assert p.reassemble_fragments == 'yes'
        assert p.receive_window_size == 109
        assert p.reset_on_timeout == 'no'
        assert p.rtt_from_client == 'yes'
        assert p.rtt_from_server == 'no'
        assert p.server_sack == 'yes'
        assert p.server_timestamp == 'no'
        assert p.syn_cookie_mss == 110
        assert p.tcp_close_timeout == 111
        assert p.tcp_generate_isn == 'yes'
        assert p.tcp_handshake_timeout == 112
        assert p.tcp_strip_sack == 'no'
        assert p.tcp_time_wait_timeout == 113
        assert p.tcp_timestamp_mode == 'rewrite'
        assert p.tcp_wscale_mode == 'strip'
        assert p.timeout_recovery == 'fallback'

    def test_api_parameters(self):
        args = load_fixture('load_ltm_fastl4_profile_1.json')
        p = ApiParameters(params=args)
        assert p.name == 'fastL4'
        assert p.description is None


class TestManager(unittest.TestCase):

    def setUp(self):
        self.spec = ArgumentSpec()

    def test_create(self, *args):
        # Configure the arguments that would be sent to the Ansible module
        set_module_args(dict(
            name='foo',
            parent='bar',
            provider=dict(
                server='localhost',
                password='password',
                user='admin'
            )
        ))

        module = AnsibleModule(
            argument_spec=self.spec.argument_spec,
            supports_check_mode=self.spec.supports_check_mode
        )
        mm = ModuleManager(module=module)

        # Override methods to force specific logic in the module to happen
        mm.exists = Mock(return_value=False)
        mm.create_on_device = Mock(return_value=True)

        results = mm.exec_module()

        assert results['changed'] is True



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2018, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'certified'}

DOCUMENTATION = r'''
---
module: bigip_device_syslog
short_description: Manage system-level syslog settings on BIG-IP
description:
  - Manage system-level syslog settings on BIG-IP.
version_added: 2.8
options:
  auth_priv_from:
    description:
      - Specifies the lowest level of messages about user authentication
        to include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  auth_priv_to:
    description:
      - Specifies the highest level of messages about user authentication
        to include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  console_log:
    description:
      - Enables or disables logging emergency syslog messages to the
        console.
    type: bool
  cron_from:
    description:
      - Specifies the lowest level of messages about time-based scheduling
        to include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  cron_to:
    description:
      - Specifies the highest level of messages about time-based
        scheduling to include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  daemon_from:
    description:
      - Specifies the lowest level of messages about daemon performance to
        include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  daemon_to:
    description:
      - Specifies the highest level of messages about daemon performance
        to include in the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  include:
    description:
      - Syslog-NG configuration to include in the device syslog config.
    type: str
  iso_date:
    description:
      - Enables or disables the ISO date format for messages in the log
        files.
    type: bool
  kern_from:
    description:
      - Specifies the lowest level of kernel messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  kern_to:
    description:
      - Specifies the highest level of kernel messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  local6_from:
    description:
      - Specifies the lowest error level for messages from the local6
        facility to include in the log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  local6_to:
    description:
      - Specifies the highest error level for messages from the local6
        facility to include in the log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  mail_from:
    description:
      - Specifies the lowest level of mail log messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  mail_to:
    description:
      - Specifies the highest level of mail log messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  messages_from:
    description:
      - Specifies the lowest level of system messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  messages_to:
    description:
      - Specifies the highest level of system messages to include in the
        system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  user_log_from:
    description:
      - Specifies the lowest level of user account messages to include in
        the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
  user_log_to:
    description:
      - Specifies the highest level of user account messages to include in
        the system log.
    type: str
    choices:
      - alert
      - crit
      - debug
      - emerg
      - err
      - info
      - notice
      - warning
extends_documentation_fragment: f5
author:
  - Tim Rupp (@caphrim007)
'''

EXAMPLES = r'''
- name: Create a syslog config
  bigip_device_syslog:
    name: foo
    provider:
      password: secret
      server: lb.mydomain.com
      user: admin
  delegate_to: localhost
'''

RETURN = r'''
auth_priv_from:
  description: The new lowest user authentication logging level
  returned: changed
  type: str
  sample: alert
auth_priv_to:
  description: The new highest user authentication logging level.
  returned: changed
  type: str
  sample: emerg
console_log:
  description: Whether logging to console is enabled or not.
  returned: changed
  type: bool
  sample: yes
iso_date:
  description: Whether ISO date format in logs is enabled or not
  returned: changed
  type: bool
  sample: no
cron_from:
  description: The new lowest time-based scheduling logging level.
  returned: changed
  type: str
  sample: emerg
cron_to:
  description: The new highest time-based scheduling logging level.
  returned: changed
  type: str
  sample: alert
daemon_from:
  description: The new lowest daemon performance logging level.
  returned: changed
  type: str
  sample: alert
daemon_to:
  description: The new highest daemon performance logging level.
  returned: changed
  type: str
  sample: alert
include:
  description: The new extra syslog-ng configuration to include in syslog config.
  returned: changed
  type: str
  sample: "filter f_remote_syslog { not (facility(local6)) };"
kern_from:
  description: The new lowest kernel messages logging level.
  returned: changed
  type: str
  sample: alert
kern_to:
  description: The new highest kernel messages logging level.
  returned: changed
  type: str
  sample: alert
local6_from:
  description: The new lowest local6 facility logging level.
  returned: changed
  type: str
  sample: alert
local6_to:
  description: The new highest local6 facility logging level.
  returned: changed
  type: str
  sample: alert
mail_from:
  description: The new lowest mail log logging level.
  returned: changed
  type: str
  sample: alert
mail_to:
  description: The new highest mail log logging level.
  returned: changed
  type: str
  sample: alert
messages_from:
  description: The new lowest system logging level.
  returned: changed
  type: str
  sample: alert
messages_to:
  description: The new highest system logging level.
  returned: changed
  type: str
  sample: alert
user_log_from:
  description: The new lowest user account logging level.
  returned: changed
  type: str
  sample: alert
user_log_to:
  description: The new highest user account logging level.
  returned: changed
  type: str
  sample: alert
'''

from ansible.module_utils.basic import AnsibleModule

try:
    from library.module_utils.network.f5.bigip import F5RestClient
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import AnsibleF5Parameters
    from library.module_utils.network.f5.common import fq_name
    from library.module_utils.network.f5.common import f5_argument_spec
    from library.module_utils.network.f5.common import flatten_boolean
    from library.module_utils.network.f5.compare import cmp_str_with_none
except ImportError:
    from ansible.module_utils.network.f5.bigip import F5RestClient
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import AnsibleF5Parameters
    from ansible.module_utils.network.f5.common import fq_name
    from ansible.module_utils.network.f5.common import f5_argument_spec
    from ansible.module_utils.network.f5.common import flatten_boolean
    from ansible.module_utils.network.f5.compare import cmp_str_with_none


class Parameters(AnsibleF5Parameters):
    api_map = {
        'authPrivFrom': 'auth_priv_from',
        'authPrivTo': 'auth_priv_to',
        'consoleLog': 'console_log',
        'cronFrom': 'cron_from',
        'cronTo': 'cron_to',
        'daemonFrom': 'daemon_from',
        'daemonTo': 'daemon_to',
        'isoDate': 'iso_date',
        'kernFrom': 'kern_from',
        'kernTo': 'kern_to',
        'local6From': 'local6_from',
        'local6To': 'local6_to',
        'mailFrom': 'mail_from',
        'mailTo': 'mail_to',
        'messagesFrom': 'messages_from',
        'messagesTo': 'messages_to',
        'userLogFrom': 'user_log_from',
        'userLogTo': 'user_log_to',
    }

    api_attributes = [
        'include',
        'authPrivFrom',
        'authPrivTo',
        'consoleLog',
        'cronFrom',
        'cronTo',
        'daemonFrom',
        'daemonTo',
        'isoDate',
        'kernFrom',
        'kernTo',
        'local6From',
        'local6To',
        'mailFrom',
        'mailTo',
        'messagesFrom',
        'messagesTo',
        'userLogFrom',
        'userLogTo',
    ]

    returnables = [
        'include',
        'auth_priv_from',
        'auth_priv_to',
        'console_log',
        'cron_from',
        'cron_to',
        'daemon_from',
        'daemon_to',
        'iso_date',
        'kern_from',
        'kern_to',
        'local6_from',
        'local6_to',
        'mail_from',
        'mail_to',
        'messages_from',
        'messages_to',
        'user_log_from',
        'user_log_to',
    ]

    updatables = [
        'include',
        'auth_priv_from',
        'auth_priv_to',
        'console_log',
        'cron_from',
        'cron_to',
        'daemon_from',
        'daemon_to',
        'iso_date',
        'kern_from',
        'kern_to',
        'local6_from',
        'local6_to',
        'mail_from',
        'mail_to',
        'messages_from',
        'messages_to',
        'user_log_from',
        'user_log_to',
    ]

    @property
    def console_log(self):
        return flatten_boolean(self._values['console_log'])

    @property
    def iso_date(self):
        return flatten_boolean(self._values['iso_date'])


class ApiParameters(Parameters):
    @property
    def include(self):
        if self._values['include'] in [None, 'none']:
            return None
        return self._values['include']


class ModuleParameters(Parameters):
    @property
    def include(self):
        if self._values['include'] is None:
            return None
        if self._values['include'] in ['', 'none']:
            return ''
        return self._values['include'].replace('"', "'")


class Changes(Parameters):
    def to_return(self):
        result = {}
        try:
            for returnable in self.returnables:
                result[returnable] = getattr(self, returnable)
            result = self._filter_params(result)
        except Exception:
            pass
        return result


class UsableChanges(Changes):
    @property
    def console_log(self):
        if self._values['console_log'] is None:
            return None
        elif self._values['console_log'] == 'yes':
            return 'enabled'
        return 'disabled'

    @property
    def iso_date(self):
        if self._values['iso_date'] is None:
            return None
        elif self._values['iso_date'] == 'yes':
            return 'enabled'
        return 'disabled'


class ReportableChanges(Changes):
    @property
    def console_log(self):
        return flatten_boolean(self._values['console_log'])

    @property
    def iso_date(self):
        return flatten_boolean(self._values['iso_date'])


class Difference(object):
    def __init__(self, want, have=None):
        self.want = want
        self.have = have

    def compare(self, param):
        try:
            result = getattr(self, param)
            return result
        except AttributeError:
            return self.__default(param)

    def __default(self, param):
        attr1 = getattr(self.want, param)
        try:
            attr2 = getattr(self.have, param)
            if attr1 != attr2:
                return attr1
        except AttributeError:
            return attr1

    @property
    def include(self):
        return cmp_str_with_none(self.want.include, self.have.include)


class ModuleManager(object):
    def __init__(self, *args, **kwargs):
        self.module = kwargs.get('module', None)
        self.client = F5RestClient(**self.module.params)
        self.want = ModuleParameters(params=self.module.params)
        self.have = ApiParameters()
        self.changes = UsableChanges()

    def _set_changed_options(self):
        changed = {}
        for key in Parameters.returnables:
            if getattr(self.want, key) is not None:
                changed[key] = getattr(self.want, key)
        if changed:
            self.changes = UsableChanges(params=changed)

    def _update_changed_options(self):
        diff = Difference(self.want, self.have)
        updatables = Parameters.updatables
        changed = dict()
        for k in updatables:
            change = diff.compare(k)
            if change is None:
                continue
            else:
                if isinstance(change, dict):
                    changed.update(change)
                else:
                    changed[k] = change
        if changed:
            self.changes = UsableChanges(params=changed)
            return True
        return False

    def should_update(self):
        result = self._update_changed_options()
        if result:
            return True
        return False

    def exec_module(self):
        result = dict()

        changed = self.present()

        reportable = ReportableChanges(params=self.changes.to_return())
        changes = reportable.to_return()
        result.update(**changes)
        result.update(dict(changed=changed))
        self._announce_deprecations(result)
        return result

    def _announce_deprecations(self, result):
        warnings = result.pop('__warnings', [])
        for warning in warnings:
            self.client.module.deprecate(
                msg=warning['msg'],
                version=warning['version']
            )

    def present(self):
        return self.update()

    def update(self):
        self.have = self.read_current_from_device()
        if not self.should_update():
            return False
        if self.module.check_mode:
            return True
        self.update_on_device()
        return True

    def update_on_device(self):
        params = self.changes.api_params()
        uri = "https://{0}:{1}/mgmt/tm/sys/syslog".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
        )
        resp = self.client.api.patch(uri, json=params)
        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)

    def read_current_from_device(self):
        uri = "https://{0}:{1}/mgmt/tm/sys/syslog".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
        )
        resp = self.client.api.get(uri)
        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)
        return ApiParameters(params=response)


class ArgumentSpec(object):
    def __init__(self):
        self.supports_check_mode = True
        levels = [
            'alert', 'crit', 'debug', 'emerg', 'err', 'info', 'notice', 'warning'
        ]

        argument_spec = dict(
            auth_priv_from=dict(choices=levels),
            auth_priv_to=dict(choices=levels),
            console_log=dict(type='bool'),
            cron_from=dict(choices=levels),
            cron_to=dict(choices=levels),
            daemon_from=dict(choices=levels),
            daemon_to=dict(choices=levels),
            include=dict(),
            iso_date=dict(type='bool'),
            kern_from=dict(choices=levels),
            kern_to=dict(choices=levels),
            local6_from=dict(choices=levels),
            local6_to=dict(choices=levels),
            mail_from=dict(choices=levels),
            mail_to=dict(choices=levels),
            messages_from=dict(choices=levels),
            messages_to=dict(choices=levels),
            user_log_from=dict(choices=levels),
            user_log_to=dict(choices=levels),
        )
        self.argument_spec = {}
        self.argument_spec.update(f5_argument_spec)
        self.argument_spec.update(argument_spec)


def main():
    spec = ArgumentSpec()

    module = AnsibleModule(
        argument_spec=spec.argument_spec,
        supports_check_mode=spec.supports_check_mode,
    )

    try:
        mm = ModuleManager(module=module)
        results = mm.exec_module()
        module.exit_json(**results)
    except F5ModuleError as ex:
        module.fail_json(msg=str(ex))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import os


try:
    from library.module_utils.network.f5.common import F5BaseClient
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import is_ansible_debug
    from library.module_utils.network.f5.icontrol import iControlRestSession
except ImportError:
    from ansible.module_utils.network.f5.common import F5BaseClient
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import is_ansible_debug
    from ansible.module_utils.network.f5.icontrol import iControlRestSession


class F5RestClient(F5BaseClient):
    def __init__(self, *args, **kwargs):
        super(F5RestClient, self).__init__(*args, **kwargs)
        self.provider = self.merge_provider_params()
        self.headers = {
            'Content-Type': 'application/json'
        }

    @property
    def api(self):
        if self._client:
            return self._client
        session, err = self.connect_via_token_auth()
        if err:
            raise F5ModuleError(err)
        self._client = session
        return session

    def connect_via_token_auth(self):
        provider = self.provider['auth_provider'] or 'local'

        url = "https://{0}:{1}/mgmt/shared/authn/login".format(
            self.provider['server'], self.provider['server_port']
        )
        payload = {
            'username': self.provider['user'],
            'password': self.provider['password'],
        }

        # - local is a special provider that is baked into the system and
        #   has no loginReference
        if provider != 'local':
            login_ref = self.get_login_ref(provider)
            payload.update(login_ref)

        session = iControlRestSession(
            validate_certs=self.provider['validate_certs']
        )

        response = session.post(
            url,
            json=payload,
            headers=self.headers
        )

        if response.status not in [200]:
            return None, response.content

        session.request.headers['X-F5-Auth-Token'] = response.json()['token']['token']
        return session, None

    def get_login_ref(self, provider):
        info = self.read_provider_info_from_device()
        uuids = [os.path.basename(os.path.dirname(x['link'])) for x in info['providers'] if '-' in x['link']]
        if provider in uuids:
            name = self.get_name_of_provider_id(info, provider)
            if not name:
                raise F5ModuleError(
                    "No name found for the provider '{0}'".format(provider)
                )
            return dict(
                loginReference=dict(
                    link="https://localhost/mgmt/cm/system/authn/providers/{0}/{1}/login".format(name, provider)
                )
            )
        names = [os.path.basename(os.path.dirname(x['link'])) for x in info['providers'] if '-' in x['link']]
        if names.count(provider) > 1:
            raise F5ModuleError(
                "Ambiguous auth_provider provided. Please specify a specific provider ID."
            )
        uuid = self.get_id_of_provider_name(info, provider)
        if not uuid:
            raise F5ModuleError(
                "No name found for the provider '{0}'".format(provider)
            )
        return dict(
            loginReference=dict(
                link="https://localhost/mgmt/cm/system/authn/providers/{0}/{1}/login".format(provider, uuid)
            )
        )

    def get_name_of_provider_id(self, info, provider):
        # Add slashes to the provider name so that it specifically finds the provider
        # as part of the URL and not a part of another substring
        provider = '/' + provider + '/'
        for x in info['providers']:
            if x['link'].find(provider) > -1:
                return x['name']
        return None

    def get_id_of_provider_name(self, info, provider):
        for x in info['providers']:
            if x['name'] == provider:
                return os.path.basename(os.path.dirname(x['link']))
        return None

    def read_provider_info_from_device(self):
        uri = "https://{0}:{1}/info/system".format(
            self.provider['server'], self.provider['server_port']
        )
        session = iControlRestSession()
        session.verify = self.provider['validate_certs']

        resp = session.get(uri)
        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)
        return response



#!/usr/bin/python
from __future__ import (absolute_import, division, print_function)
# Copyright 2019 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__metaclass__ = type

ANSIBLE_METADATA = {'status': ['preview'],
                    'supported_by': 'community',
                    'metadata_version': '1.1'}

DOCUMENTATION = '''
---
module: fortios_system_virtual_wan_link
short_description: Configure redundant internet connections using SD-WAN (formerly virtual WAN link) in Fortinet's FortiOS and FortiGate.
description:
    - This module is able to configure a FortiGate or FortiOS by allowing the
      user to set and modify system feature and virtual_wan_link category.
      Examples include all parameters and values need to be adjusted to datasources before usage.
      Tested with FOS v6.0.2
version_added: "2.8"
author:
    - Miguel Angel Munoz (@mamunozgonzalez)
    - Nicolas Thomas (@thomnico)
notes:
    - Requires fortiosapi library developed by Fortinet
    - Run as a local_action in your playbook
requirements:
    - fortiosapi>=0.9.8
options:
    host:
       description:
            - FortiOS or FortiGate ip address.
       required: true
    username:
        description:
            - FortiOS or FortiGate username.
        required: true
    password:
        description:
            - FortiOS or FortiGate password.
        default: ""
    vdom:
        description:
            - Virtual domain, among those defined previously. A vdom is a
              virtual instance of the FortiGate that can be configured and
              used as a different unit.
        default: root
    https:
        description:
            - Indicates if the requests towards FortiGate must use HTTPS
              protocol
        type: bool
        default: true
    system_virtual_wan_link:
        description:
            - Configure redundant internet connections using SD-WAN (formerly virtual WAN link).
        default: null
        suboptions:
            fail-alert-interfaces:
                description:
                    - Physical interfaces that will be alerted.
                suboptions:
                    name:
                        description:
                            - Physical interface name. Source system.interface.name.
                        required: true
            fail-detect:
                description:
                    - Enable/disable SD-WAN Internet connection status checking (failure detection).
                choices:
                    - enable
                    - disable
            health-check:
                description:
                    - SD-WAN status checking or health checking. Identify a server on the Internet and determine how SD-WAN verifies that the FortiGate can
                       communicate with it.
                suboptions:
                    addr-mode:
                        description:
                            - Address mode (IPv4 or IPv6).
                        choices:
                            - ipv4
                            - ipv6
                    failtime:
                        description:
                            - Number of failures before server is considered lost (1 - 10, default = 5).
                    http-get:
                        description:
                            - URL used to communicate with the server if the protocol if the protocol is HTTP.
                    http-match:
                        description:
                            - Response string expected from the server if the protocol is HTTP.
                    interval:
                        description:
                            - Status check interval, or the time between attempting to connect to the server (1 - 3600 sec, default = 5).
                    members:
                        description:
                            - Member sequence number list.
                        suboptions:
                            seq-num:
                                description:
                                    - Member sequence number. Source system.virtual-wan-link.members.seq-num.
                                required: true
                    name:
                        description:
                            - Status check or health check name.
                        required: true
                    packet-size:
                        description:
                            - Packet size of a twamp test session,
                    password:
                        description:
                            - Twamp controller password in authentication mode
                    port:
                        description:
                            - Port number used to communicate with the server over the selected protocol.
                    protocol:
                        description:
                            - Protocol used to determine if the FortiGate can communicate with the server.
                        choices:
                            - ping
                            - tcp-echo
                            - udp-echo
                            - http
                            - twamp
                            - ping6
                    recoverytime:
                        description:
                            - Number of successful responses received before server is considered recovered (1 - 10, default = 5).
                    security-mode:
                        description:
                            - Twamp controller security mode.
                        choices:
                            - none
                            - authentication
                    server:
                        description:
                            - IP address or FQDN name of the server.
                    sla:
                        description:
                            - Service level agreement (SLA).
                        suboptions:
                            id:
                                description:
                                    - SLA ID.
                                required: true
                            jitter-threshold:
                                description:
                                    - Jitter for SLA to make decision in milliseconds. (0 - 10000000, default = 5).
                            latency-threshold:
                                description:
                                    - Latency for SLA to make decision in milliseconds. (0 - 10000000, default = 5).
                            link-cost-factor:
                                description:
                                    - Criteria on which to base link selection.
                                choices:
                                    - latency
                                    - jitter
                                    - packet-loss
                            packetloss-threshold:
                                description:
                                    - Packet loss for SLA to make decision in percentage. (0 - 100, default = 0).
                    threshold-alert-jitter:
                        description:
                            - Alert threshold for jitter (ms, default = 0).
                    threshold-alert-latency:
                        description:
                            - Alert threshold for latency (ms, default = 0).
                    threshold-alert-packetloss:
                        description:
                            - Alert threshold for packet loss (percentage, default = 0).
                    threshold-warning-jitter:
                        description:
                            - Warning threshold for jitter (ms, default = 0).
                    threshold-warning-latency:
                        description:
                            - Warning threshold for latency (ms, default = 0).
                    threshold-warning-packetloss:
                        description:
                            - Warning threshold for packet loss (percentage, default = 0).
                    update-cascade-interface:
                        description:
                            - Enable/disable update cascade interface.
                        choices:
                            - enable
                            - disable
                    update-static-route:
                        description:
                            - Enable/disable updating the static route.
                        choices:
                            - enable
                            - disable
            load-balance-mode:
                description:
                    - Algorithm or mode to use for load balancing Internet traffic to SD-WAN members.
                choices:
                    - source-ip-based
                    - weight-based
                    - usage-based
                    - source-dest-ip-based
                    - measured-volume-based
            members:
                description:
                    - Physical FortiGate interfaces added to the virtual-wan-link.
                suboptions:
                    comment:
                        description:
                            - Comments.
                    gateway:
                        description:
                            - The default gateway for this interface. Usually the default gateway of the Internet service provider that this interface is
                               connected to.
                    gateway6:
                        description:
                            - IPv6 gateway.
                    ingress-spillover-threshold:
                        description:
                            - Ingress spillover threshold for this interface (0 - 16776000 kbit/s). When this traffic volume threshold is reached, new
                               sessions spill over to other interfaces in the SD-WAN.
                    interface:
                        description:
                            - Interface name. Source system.interface.name.
                    priority:
                        description:
                            - Priority of the interface (0 - 4294967295). Used for SD-WAN rules or priority rules.
                    seq-num:
                        description:
                            - Sequence number(1-255).
                        required: true
                    source:
                        description:
                            - Source IP address used in the health-check packet to the server.
                    source6:
                        description:
                            - Source IPv6 address used in the health-check packet to the server.
                    spillover-threshold:
                        description:
                            - Egress spillover threshold for this interface (0 - 16776000 kbit/s). When this traffic volume threshold is reached, new sessions
                               spill over to other interfaces in the SD-WAN.
                    status:
                        description:
                            - Enable/disable this interface in the SD-WAN.
                        choices:
                            - disable
                            - enable
                    volume-ratio:
                        description:
                            - Measured volume ratio (this value / sum of all values = percentage of link volume, 0 - 255).
                    weight:
                        description:
                            - Weight of this interface for weighted load balancing. (0 - 255) More traffic is directed to interfaces with higher weights.
            service:
                description:
                    - Create SD-WAN rules or priority rules (also called services) to control how sessions are distributed to physical interfaces in the
                       SD-WAN.
                suboptions:
                    addr-mode:
                        description:
                            - Address mode (IPv4 or IPv6).
                        choices:
                            - ipv4
                            - ipv6
                    bandwidth-weight:
                        description:
                            - Coefficient of reciprocal of available bidirectional bandwidth in the formula of custom-profile-1.
                    dscp-forward:
                        description:
                            - Enable/disable forward traffic DSCP tag.
                        choices:
                            - enable
                            - disable
                    dscp-forward-tag:
                        description:
                            - Forward traffic DSCP tag.
                    dscp-reverse:
                        description:
                            - Enable/disable reverse traffic DSCP tag.
                        choices:
                            - enable
                            - disable
                    dscp-reverse-tag:
                        description:
                            - Reverse traffic DSCP tag.
                    dst:
                        description:
                            - Destination address name.
                        suboptions:
                            name:
                                description:
                                    - Address or address group name. Source firewall.address.name firewall.addrgrp.name.
                                required: true
                    dst-negate:
                        description:
                            - Enable/disable negation of destination address match.
                        choices:
                            - enable
                            - disable
                    dst6:
                        description:
                            - Destination address6 name.
                        suboptions:
                            name:
                                description:
                                    - Address6 or address6 group name. Source firewall.address6.name firewall.addrgrp6.name.
                                required: true
                    end-port:
                        description:
                            - End destination port number.
                    gateway:
                        description:
                            - Enable/disable SD-WAN service gateway.
                        choices:
                            - enable
                            - disable
                    groups:
                        description:
                            - User groups.
                        suboptions:
                            name:
                                description:
                                    - Group name. Source user.group.name.
                                required: true
                    health-check:
                        description:
                            - Health check. Source system.virtual-wan-link.health-check.name.
                    hold-down-time:
                        description:
                            - Waiting period in seconds when switching from the back-up member to the primary member (0 - 10000000, default = 0).
                    id:
                        description:
                            - Priority rule ID (1 - 4000).
                        required: true
                    input-device:
                        description:
                            - Source interface name.
                        suboptions:
                            name:
                                description:
                                    - Interface name. Source system.interface.name.
                                required: true
                    internet-service:
                        description:
                            - Enable/disable use of Internet service for application-based load balancing.
                        choices:
                            - enable
                            - disable
                    internet-service-ctrl:
                        description:
                            - Control-based Internet Service ID list.
                        suboptions:
                            id:
                                description:
                                    - Control-based Internet Service ID.
                                required: true
                    internet-service-ctrl-group:
                        description:
                            - Control-based Internet Service group list.
                        suboptions:
                            name:
                                description:
                                    - Control-based Internet Service group name. Source application.group.name.
                                required: true
                    internet-service-custom:
                        description:
                            - Custom Internet service name list.
                        suboptions:
                            name:
                                description:
                                    - Custom Internet service name. Source firewall.internet-service-custom.name.
                                required: true
                    internet-service-custom-group:
                        description:
                            - Custom Internet Service group list.
                        suboptions:
                            name:
                                description:
                                    - Custom Internet Service group name. Source firewall.internet-service-custom-group.name.
                                required: true
                    internet-service-group:
                        description:
                            - Internet Service group list.
                        suboptions:
                            name:
                                description:
                                    - Internet Service group name. Source firewall.internet-service-group.name.
                                required: true
                    internet-service-id:
                        description:
                            - Internet service ID list.
                        suboptions:
                            id:
                                description:
                                    - Internet service ID. Source firewall.internet-service.id.
                                required: true
                    jitter-weight:
                        description:
                            - Coefficient of jitter in the formula of custom-profile-1.
                    latency-weight:
                        description:
                            - Coefficient of latency in the formula of custom-profile-1.
                    link-cost-factor:
                        description:
                            - Link cost factor.
                        choices:
                            - latency
                            - jitter
                            - packet-loss
                            - inbandwidth
                            - outbandwidth
                            - bibandwidth
                            - custom-profile-1
                    link-cost-threshold:
                        description:
                            - Percentage threshold change of link cost values that will result in policy route regeneration (0 - 10000000, default = 10).
                    member:
                        description:
                            - Member sequence number.
                    mode:
                        description:
                            - Control how the priority rule sets the priority of interfaces in the SD-WAN.
                        choices:
                            - auto
                            - manual
                            - priority
                            - sla
                    name:
                        description:
                            - Priority rule name.
                    packet-loss-weight:
                        description:
                            - Coefficient of packet-loss in the formula of custom-profile-1.
                    priority-members:
                        description:
                            - Member sequence number list.
                        suboptions:
                            seq-num:
                                description:
                                    - Member sequence number. Source system.virtual-wan-link.members.seq-num.
                                required: true
                    protocol:
                        description:
                            - Protocol number.
                    quality-link:
                        description:
                            - Quality grade.
                    route-tag:
                        description:
                            - IPv4 route map route-tag.
                    sla:
                        description:
                            - Service level agreement (SLA).
                        suboptions:
                            health-check:
                                description:
                                    - Virtual WAN Link health-check. Source system.virtual-wan-link.health-check.name.
                                required: true
                            id:
                                description:
                                    - SLA ID.
                    src:
                        description:
                            - Source address name.
                        suboptions:
                            name:
                                description:
                                    - Address or address group name. Source firewall.address.name firewall.addrgrp.name.
                                required: true
                    src-negate:
                        description:
                            - Enable/disable negation of source address match.
                        choices:
                            - enable
                            - disable
                    src6:
                        description:
                            - Source address6 name.
                        suboptions:
                            name:
                                description:
                                    - Address6 or address6 group name. Source firewall.address6.name firewall.addrgrp6.name.
                                required: true
                    start-port:
                        description:
                            - Start destination port number.
                    status:
                        description:
                            - Enable/disable SD-WAN service.
                        choices:
                            - enable
                            - disable
                    tos:
                        description:
                            - Type of service bit pattern.
                    tos-mask:
                        description:
                            - Type of service evaluated bits.
                    users:
                        description:
                            - User name.
                        suboptions:
                            name:
                                description:
                                    - User name. Source user.local.name.
                                required: true
            status:
                description:
                    - Enable/disable SD-WAN.
                choices:
                    - disable
                    - enable
'''

EXAMPLES = '''
- hosts: localhost
  vars:
   host: "192.168.122.40"
   username: "admin"
   password: ""
   vdom: "root"
  tasks:
  - name: Configure redundant internet connections using SD-WAN (formerly virtual WAN link).
    fortios_system_virtual_wan_link:
      host:  "{{ host }}"
      username: "{{ username }}"
      password: "{{ password }}"
      vdom:  "{{ vdom }}"
      https: "False"
      system_virtual_wan_link:
        fail-alert-interfaces:
         -
            name: "default_name_4 (source system.interface.name)"
        fail-detect: "enable"
        health-check:
         -
            addr-mode: "ipv4"
            failtime: "8"
            http-get: "<your_own_value>"
            http-match: "<your_own_value>"
            interval: "11"
            members:
             -
                seq-num: "13 (source system.virtual-wan-link.members.seq-num)"
            name: "default_name_14"
            packet-size: "15"
            password: "<your_own_value>"
            port: "17"
            protocol: "ping"
            recoverytime: "19"
            security-mode: "none"
            server: "192.168.100.40"
            sla:
             -
                id:  "23"
                jitter-threshold: "24"
                latency-threshold: "25"
                link-cost-factor: "latency"
                packetloss-threshold: "27"
            threshold-alert-jitter: "28"
            threshold-alert-latency: "29"
            threshold-alert-packetloss: "30"
            threshold-warning-jitter: "31"
            threshold-warning-latency: "32"
            threshold-warning-packetloss: "33"
            update-cascade-interface: "enable"
            update-static-route: "enable"
        load-balance-mode: "source-ip-based"
        members:
         -
            comment: "Comments."
            gateway: "<your_own_value>"
            gateway6: "<your_own_value>"
            ingress-spillover-threshold: "41"
            interface: "<your_own_value> (source system.interface.name)"
            priority: "43"
            seq-num: "44"
            source: "<your_own_value>"
            source6: "<your_own_value>"
            spillover-threshold: "47"
            status: "disable"
            volume-ratio: "49"
            weight: "50"
        service:
         -
            addr-mode: "ipv4"
            bandwidth-weight: "53"
            dscp-forward: "enable"
            dscp-forward-tag: "<your_own_value>"
            dscp-reverse: "enable"
            dscp-reverse-tag: "<your_own_value>"
            dst:
             -
                name: "default_name_59 (source firewall.address.name firewall.addrgrp.name)"
            dst-negate: "enable"
            dst6:
             -
                name: "default_name_62 (source firewall.address6.name firewall.addrgrp6.name)"
            end-port: "63"
            gateway: "enable"
            groups:
             -
                name: "default_name_66 (source user.group.name)"
            health-check: "<your_own_value> (source system.virtual-wan-link.health-check.name)"
            hold-down-time: "68"
            id:  "69"
            input-device:
             -
                name: "default_name_71 (source system.interface.name)"
            internet-service: "enable"
            internet-service-ctrl:
             -
                id:  "74"
            internet-service-ctrl-group:
             -
                name: "default_name_76 (source application.group.name)"
            internet-service-custom:
             -
                name: "default_name_78 (source firewall.internet-service-custom.name)"
            internet-service-custom-group:
             -
                name: "default_name_80 (source firewall.internet-service-custom-group.name)"
            internet-service-group:
             -
                name: "default_name_82 (source firewall.internet-service-group.name)"
            internet-service-id:
             -
                id:  "84 (source firewall.internet-service.id)"
            jitter-weight: "85"
            latency-weight: "86"
            link-cost-factor: "latency"
            link-cost-threshold: "88"
            member: "89"
            mode: "auto"
            name: "default_name_91"
            packet-loss-weight: "92"
            priority-members:
             -
                seq-num: "94 (source system.virtual-wan-link.members.seq-num)"
            protocol: "95"
            quality-link: "96"
            route-tag: "97"
            sla:
             -
                health-check: "<your_own_value> (source system.virtual-wan-link.health-check.name)"
                id:  "100"
            src:
             -
                name: "default_name_102 (source firewall.address.name firewall.addrgrp.name)"
            src-negate: "enable"
            src6:
             -
                name: "default_name_105 (source firewall.address6.name firewall.addrgrp6.name)"
            start-port: "106"
            status: "enable"
            tos: "<your_own_value>"
            tos-mask: "<your_own_value>"
            users:
             -
                name: "default_name_111 (source user.local.name)"
        status: "disable"
'''

RETURN = '''
build:
  description: Build number of the fortigate image
  returned: always
  type: str
  sample: '1547'
http_method:
  description: Last method used to provision the content into FortiGate
  returned: always
  type: str
  sample: 'PUT'
http_status:
  description: Last result given by FortiGate on last operation applied
  returned: always
  type: str
  sample: "200"
mkey:
  description: Master key (id) used in the last call to FortiGate
  returned: success
  type: str
  sample: "id"
name:
  description: Name of the table used to fulfill the request
  returned: always
  type: str
  sample: "urlfilter"
path:
  description: Path of the table used to fulfill the request
  returned: always
  type: str
  sample: "webfilter"
revision:
  description: Internal revision number
  returned: always
  type: str
  sample: "17.0.2.10658"
serial:
  description: Serial number of the unit
  returned: always
  type: str
  sample: "FGVMEVYYQT3AB5352"
status:
  description: Indication of the operation's result
  returned: always
  type: str
  sample: "success"
vdom:
  description: Virtual domain used
  returned: always
  type: str
  sample: "root"
version:
  description: Version of the FortiGate
  returned: always
  type: str
  sample: "v5.6.3"

'''

from ansible.module_utils.basic import AnsibleModule


def login(data, fos):
    host = data['host']
    username = data['username']
    password = data['password']

    fos.debug('on')
    if 'https' in data and not data['https']:
        fos.https('off')
    else:
        fos.https('on')

    fos.login(host, username, password)


def filter_system_virtual_wan_link_data(json):
    option_list = ['fail-alert-interfaces', 'fail-detect', 'health-check',
                   'load-balance-mode', 'members', 'service',
                   'status']
    dictionary = {}

    for attribute in option_list:
        if attribute in json and json[attribute] is not None:
            dictionary[attribute] = json[attribute]

    return dictionary


def system_virtual_wan_link(data, fos):
    vdom = data['vdom']
    system_virtual_wan_link_data = data['system_virtual_wan_link']
    filtered_data = filter_system_virtual_wan_link_data(system_virtual_wan_link_data)

    return fos.set('system',
                   'virtual-wan-link',
                   data=filtered_data,
                   vdom=vdom)


def fortios_system(data, fos):
    login(data, fos)

    if data['system_virtual_wan_link']:
        resp = system_virtual_wan_link(data, fos)

    fos.logout()
    return not resp['status'] == "success", resp['status'] == "success", resp


def main():
    fields = {
        "host": {"required": True, "type": "str"},
        "username": {"required": True, "type": "str"},
        "password": {"required": False, "type": "str", "no_log": True},
        "vdom": {"required": False, "type": "str", "default": "root"},
        "https": {"required": False, "type": "bool", "default": True},
        "system_virtual_wan_link": {
            "required": False, "type": "dict",
            "options": {
                "fail-alert-interfaces": {"required": False, "type": "list",
                                          "options": {
                                              "name": {"required": True, "type": "str"}
                                          }},
                "fail-detect": {"required": False, "type": "str",
                                "choices": ["enable", "disable"]},
                "health-check": {"required": False, "type": "list",
                                 "options": {
                                     "addr-mode": {"required": False, "type": "str",
                                                   "choices": ["ipv4", "ipv6"]},
                                     "failtime": {"required": False, "type": "int"},
                                     "http-get": {"required": False, "type": "str"},
                                     "http-match": {"required": False, "type": "str"},
                                     "interval": {"required": False, "type": "int"},
                                     "members": {"required": False, "type": "list",
                                                 "options": {
                                                     "seq-num": {"required": True, "type": "int"}
                                                 }},
                                     "name": {"required": True, "type": "str"},
                                     "packet-size": {"required": False, "type": "int"},
                                     "password": {"required": False, "type": "str"},
                                     "port": {"required": False, "type": "int"},
                                     "protocol": {"required": False, "type": "str",
                                                  "choices": ["ping", "tcp-echo", "udp-echo",
                                                              "http", "twamp", "ping6"]},
                                     "recoverytime": {"required": False, "type": "int"},
                                     "security-mode": {"required": False, "type": "str",
                                                       "choices": ["none", "authentication"]},
                                     "server": {"required": False, "type": "str"},
                                     "sla": {"required": False, "type": "list",
                                             "options": {
                                                 "id": {"required": True, "type": "int"},
                                                 "jitter-threshold": {"required": False, "type": "int"},
                                                 "latency-threshold": {"required": False, "type": "int"},
                                                 "link-cost-factor": {"required": False, "type": "str",
                                                                      "choices": ["latency", "jitter", "packet-loss"]},
                                                 "packetloss-threshold": {"required": False, "type": "int"}
                                             }},
                                     "threshold-alert-jitter": {"required": False, "type": "int"},
                                     "threshold-alert-latency": {"required": False, "type": "int"},
                                     "threshold-alert-packetloss": {"required": False, "type": "int"},
                                     "threshold-warning-jitter": {"required": False, "type": "int"},
                                     "threshold-warning-latency": {"required": False, "type": "int"},
                                     "threshold-warning-packetloss": {"required": False, "type": "int"},
                                     "update-cascade-interface": {"required": False, "type": "str",
                                                                  "choices": ["enable", "disable"]},
                                     "update-static-route": {"required": False, "type": "str",
                                                             "choices": ["enable", "disable"]}
                                 }},
                "load-balance-mode": {"required": False, "type": "str",
                                      "choices": ["source-ip-based", "weight-based", "usage-based",
                                                  "source-dest-ip-based", "measured-volume-based"]},
                "members": {"required": False, "type": "list",
                            "options": {
                                "comment": {"required": False, "type": "str"},
                                "gateway": {"required": False, "type": "str"},
                                "gateway6": {"required": False, "type": "str"},
                                "ingress-spillover-threshold": {"required": False, "type": "int"},
                                "interface": {"required": False, "type": "str"},
                                "priority": {"required": False, "type": "int"},
                                "seq-num": {"required": True, "type": "int"},
                                "source": {"required": False, "type": "str"},
                                "source6": {"required": False, "type": "str"},
                                "spillover-threshold": {"required": False, "type": "int"},
                                "status": {"required": False, "type": "str",
                                           "choices": ["disable", "enable"]},
                                "volume-ratio": {"required": False, "type": "int"},
                                "weight": {"required": False, "type": "int"}
                            }},
                "service": {"required": False, "type": "list",
                            "options": {
                                "addr-mode": {"required": False, "type": "str",
                                              "choices": ["ipv4", "ipv6"]},
                                "bandwidth-weight": {"required": False, "type": "int"},
                                "dscp-forward": {"required": False, "type": "str",
                                                 "choices": ["enable", "disable"]},
                                "dscp-forward-tag": {"required": False, "type": "str"},
                                "dscp-reverse": {"required": False, "type": "str",
                                                 "choices": ["enable", "disable"]},
                                "dscp-reverse-tag": {"required": False, "type": "str"},
                                "dst": {"required": False, "type": "list",
                                        "options": {
                                            "name": {"required": True, "type": "str"}
                                        }},
                                "dst-negate": {"required": False, "type": "str",
                                               "choices": ["enable", "disable"]},
                                "dst6": {"required": False, "type": "list",
                                         "options": {
                                             "name": {"required": True, "type": "str"}
                                         }},
                                "end-port": {"required": False, "type": "int"},
                                "gateway": {"required": False, "type": "str",
                                            "choices": ["enable", "disable"]},
                                "groups": {"required": False, "type": "list",
                                           "options": {
                                               "name": {"required": True, "type": "str"}
                                           }},
                                "health-check": {"required": False, "type": "str"},
                                "hold-down-time": {"required": False, "type": "int"},
                                "id": {"required": True, "type": "int"},
                                "input-device": {"required": False, "type": "list",
                                                 "options": {
                                                     "name": {"required": True, "type": "str"}
                                                 }},
                                "internet-service": {"required": False, "type": "str",
                                                     "choices": ["enable", "disable"]},
                                "internet-service-ctrl": {"required": False, "type": "list",
                                                          "options": {
                                                              "id": {"required": True, "type": "int"}
                                                          }},
                                "internet-service-ctrl-group": {"required": False, "type": "list",
                                                                "options": {
                                                                    "name": {"required": True, "type": "str"}
                                                                }},
                                "internet-service-custom": {"required": False, "type": "list",
                                                            "options": {
                                                                "name": {"required": True, "type": "str"}
                                                            }},
                                "internet-service-custom-group": {"required": False, "type": "list",
                                                                  "options": {
                                                                      "name": {"required": True, "type": "str"}
                                                                  }},
                                "internet-service-group": {"required": False, "type": "list",
                                                           "options": {
                                                               "name": {"required": True, "type": "str"}
                                                           }},
                                "internet-service-id": {"required": False, "type": "list",
                                                        "options": {
                                                            "id": {"required": True, "type": "int"}
                                                        }},
                                "jitter-weight": {"required": False, "type": "int"},
                                "latency-weight": {"required": False, "type": "int"},
                                "link-cost-factor": {"required": False, "type": "str",
                                                     "choices": ["latency", "jitter", "packet-loss",
                                                                 "inbandwidth", "outbandwidth", "bibandwidth",
                                                                 "custom-profile-1"]},
                                "link-cost-threshold": {"required": False, "type": "int"},
                                "member": {"required": False, "type": "int"},
                                "mode": {"required": False, "type": "str",
                                         "choices": ["auto", "manual", "priority",
                                                     "sla"]},
                                "name": {"required": False, "type": "str"},
                                "packet-loss-weight": {"required": False, "type": "int"},
                                "priority-members": {"required": False, "type": "list",
                                                     "options": {
                                                         "seq-num": {"required": True, "type": "int"}
                                                     }},
                                "protocol": {"required": False, "type": "int"},
                                "quality-link": {"required": False, "type": "int"},
                                "route-tag": {"required": False, "type": "int"},
                                "sla": {"required": False, "type": "list",
                                        "options": {
                                            "health-check": {"required": True, "type": "str"},
                                            "id": {"required": False, "type": "int"}
                                        }},
                                "src": {"required": False, "type": "list",
                                        "options": {
                                            "name": {"required": True, "type": "str"}
                                        }},
                                "src-negate": {"required": False, "type": "str",
                                               "choices": ["enable", "disable"]},
                                "src6": {"required": False, "type": "list",
                                         "options": {
                                             "name": {"required": True, "type": "str"}
                                         }},
                                "start-port": {"required": False, "type": "int"},
                                "status": {"required": False, "type": "str",
                                           "choices": ["enable", "disable"]},
                                "tos": {"required": False, "type": "str"},
                                "tos-mask": {"required": False, "type": "str"},
                                "users": {"required": False, "type": "list",
                                          "options": {
                                              "name": {"required": True, "type": "str"}
                                          }}
                            }},
                "status": {"required": False, "type": "str",
                           "choices": ["disable", "enable"]}

            }
        }
    }

    module = AnsibleModule(argument_spec=fields,
                           supports_check_mode=False)
    try:
        from fortiosapi import FortiOSAPI
    except ImportError:
        module.fail_json(msg="fortiosapi module is required")

    fos = FortiOSAPI()

    is_error, has_changed, result = fortios_system(module.params, fos)

    if not is_error:
        module.exit_json(changed=has_changed, meta=result)
    else:
        module.fail_json(msg="Error in repo", meta=result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2018 Catalyst IT Ltd.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: os_coe_cluster
short_description: Add/Remove COE cluster from OpenStack Cloud
extends_documentation_fragment: openstack
version_added: "2.8"
author: "Feilong Wang (@flwang)"
description:
   - Add or Remove COE cluster from the OpenStack Container Infra service.
options:
   availability_zone:
     description:
       - Ignored. Present for backwards compatibility
   cluster_template_id:
     description:
        - The template ID of cluster template.
     required: true
   discovery_url:
       description:
         - Url used for cluster node discovery
   docker_volume_size:
      description:
         - The size in GB of the docker volume
   flavor_id:
      description:
         - The flavor of the minion node for this ClusterTemplate
   keypair:
      description:
         - Name of the keypair to use.
   labels:
      description:
         - One or more key/value pairs
   master_flavor_id:
      description:
         - The flavor of the master node for this ClusterTemplate
   master_count:
      description:
         - The number of master nodes for this cluster
      default: 1
   name:
      description:
         - Name that has to be given to the cluster template
      required: true
   node_count:
      description:
         - The number of nodes for this cluster
      default: 1
   state:
      description:
         - Indicate desired state of the resource.
      choices: [present, absent]
      default: present
   timeout:
      description:
         - Timeout for creating the cluster in minutes. Default to 60 mins
           if not set
      default: 60
requirements: ["openstacksdk"]
'''

RETURN = '''
id:
    description: The cluster UUID.
    returned: On success when I(state) is 'present'
    type: str
    sample: "39007a7e-ee4f-4d13-8283-b4da2e037c69"
cluster:
    description: Dictionary describing the cluster.
    returned: On success when I(state) is 'present'
    type: complex
    contains:
      api_address:
          description:
            - Api address of cluster master node
          type: string
          sample: https://172.24.4.30:6443
      cluster_template_id:
          description: The cluster_template UUID
          type: string
          sample: '7b1418c8-cea8-48fc-995d-52b66af9a9aa'
      coe_version:
          description:
            - Version of the COE software currently running in this cluster
          type: string
          sample: v1.11.1
      container_version:
          description:
            - Version of the container software. Example: docker version.
          type: string
          sample: 1.12.6
      created_at:
          description:
            - The time in UTC at which the cluster is created
          type: datetime
          sample: 2018-08-16T10:29:45+00:00
      create_timeout:
          description:
            - Timeout for creating the cluster in minutes. Default to 60 if
              not set.
          type: int
          sample: 60
      discovery_url:
          description:
            - Url used for cluster node discovery
          type: string
          sample: https://discovery.etcd.io/a42ee38e7113f31f4d6324f24367aae5
      faults:
          description:
            - Fault info collected from the Heat resources of this cluster
          type: dict
          sample: {'0': 'ResourceInError: resources[0].resources...'}
      flavor_id:
          description:
            - The flavor of the minion node for this cluster
          type: string
          sample: c1.c1r1
      keypair:
          description:
            - Name of the keypair to use.
          type: string
          sample: mykey
      labels:
          description: One or more key/value pairs
          type: dict
          sample: {'key1': 'value1', 'key2': 'value2'}
      master_addresses:
          description:
            - IP addresses of cluster master nodes
          type: list
          sample: ['172.24.4.5']
      master_count:
          description:
            - The number of master nodes for this cluster.
          type: int
          sample: 1
      master_flavor_id:
          description:
            - The flavor of the master node for this cluster
          type: string
          sample: c1.c1r1
      name:
          description:
            - Name that has to be given to the cluster
          type: string
          sample: k8scluster
      node_addresses:
          description:
            - IP addresses of cluster slave nodes
          type: list
          sample: ['172.24.4.8']
      node_count:
          description:
            - The number of master nodes for this cluster.
          type: int
          sample: 1
      stack_id:
          description:
            - Stack id of the Heat stack
          type: string
          sample: '07767ec6-85f5-44cb-bd63-242a8e7f0d9d'
      status:
          description: Status of the cluster from the heat stack
          type: string
          sample: 'CREATE_COMLETE'
      status_reason:
          description:
            - Status reason of the cluster from the heat stack
          type: string
          sample: 'Stack CREATE completed successfully'
      updated_at:
          description:
            - The time in UTC at which the cluster is updated
          type: datetime
          sample: '2018-08-16T10:39:25+00:00'
      uuid:
          description:
            - Unique UUID for this cluster
          type: string
          sample: '86246a4d-a16c-4a58-9e96ad7719fe0f9d'
'''

EXAMPLES = '''
# Create a new Kubernetes cluster
- os_coe_cluster:
    name: k8s
    cluster_template_id: k8s-ha
    keypair: mykey
    master_count: 3
    node_count: 5
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.openstack import openstack_full_argument_spec, openstack_module_kwargs, openstack_cloud_from_module


def _parse_labels(labels):
    if isinstance(labels, str):
        labels_dict = {}
        for kv_str in labels.split(","):
            k, v = kv_str.split("=")
            labels_dict[k] = v
        return labels_dict
    if not labels:
        return {}
    return labels


def main():
    argument_spec = openstack_full_argument_spec(
        cluster_template_id=dict(required=True),
        discovery_url=dict(default=None),
        docker_volume_size=dict(type='int'),
        flavor_id=dict(default=None),
        keypair=dict(default=None),
        labels=dict(default=None, type='raw'),
        master_count=dict(type='int', default=1),
        master_flavor_id=dict(default=None),
        name=dict(required=True),
        node_count=dict(type='int', default=1),
        state=dict(default='present', choices=['absent', 'present']),
        timeout=dict(type='int', default=60),
    )
    module_kwargs = openstack_module_kwargs()
    module = AnsibleModule(argument_spec, **module_kwargs)

    params = module.params.copy()

    state = module.params['state']
    name = module.params['name']
    cluster_template_id = module.params['cluster_template_id']

    kwargs = dict(
        discovery_url=module.params['discovery_url'],
        docker_volume_size=module.params['docker_volume_size'],
        flavor_id=module.params['flavor_id'],
        keypair=module.params['keypair'],
        labels=_parse_labels(params['labels']),
        master_count=module.params['master_count'],
        master_flavor_id=module.params['master_flavor_id'],
        node_count=module.params['node_count'],
        create_timeout=module.params['timeout'],
    )

    sdk, cloud = openstack_cloud_from_module(module)
    try:
        changed = False
        cluster = cloud.get_coe_cluster(name_or_id=name, filters={'cluster_template_id': cluster_template_id})

        if state == 'present':
            if not cluster:
                cluster = cloud.create_coe_cluster(name, cluster_template_id=cluster_template_id, **kwargs)
                changed = True
            else:
                changed = False

            module.exit_json(changed=changed, cluster=cluster, id=cluster['uuid'])
        elif state == 'absent':
            if not cluster:
                module.exit_json(changed=False)
            else:
                cloud.delete_coe_cluster(name)
                module.exit_json(changed=True)
    except sdk.exceptions.OpenStackCloudException as e:
        module.fail_json(msg=str(e), extra_data=e.extra_data)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2017, F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['stableinterface'],
                    'supported_by': 'certified'}

DOCUMENTATION = r'''
---
module: bigip_iapp_service
short_description: Manages TCL iApp services on a BIG-IP
description:
  - Manages TCL iApp services on a BIG-IP.
  - If you are looking for the API that is communicated with on the BIG-IP,
    the one the is used is C(/mgmt/tm/sys/application/service/).
version_added: 2.4
options:
  name:
    description:
      - The name of the iApp service that you want to deploy.
    type: str
    required: True
  template:
    description:
      - The iApp template from which to instantiate a new service. This
        template must exist on your BIG-IP before you can successfully
        create a service.
      - When creating a new service, this parameter is required.
    type: str
  parameters:
    description:
      - A hash of all the required template variables for the iApp template.
        If your parameters are stored in a file (the more common scenario)
        it is recommended you use either the C(file) or C(template) lookups
        to supply the expected parameters.
      - These parameters typically consist of the C(lists), C(tables), and
        C(variables) fields.
    type: dict
  force:
    description:
      - Forces the updating of an iApp service even if the parameters to the
        service have not changed. This option is of particular importance if
        the iApp template that underlies the service has been updated in-place.
        This option is equivalent to re-configuring the iApp if that template
        has changed.
    type: bool
    default: no
  state:
    description:
      - When C(present), ensures that the iApp service is created and running.
        When C(absent), ensures that the iApp service has been removed.
    type: str
    choices:
      - present
      - absent
    default: present
  partition:
    description:
      - Device partition to manage resources on.
    type: str
    default: Common
    version_added: 2.5
  strict_updates:
    description:
      - Indicates whether the application service is tied to the template,
        so when the template is updated, the application service changes to
        reflect the updates.
      - When C(yes), disallows any updates to the resources that the iApp
        service has created, if they are not updated directly through the
        iApp.
      - When C(no), allows updates outside of the iApp.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Service payload that you provide in
        the C(parameters) field.
    type: bool
    default: yes
    version_added: 2.5
  traffic_group:
    description:
      - The traffic group for the iApp service. When creating a new service, if
        this value is not specified, the default of C(/Common/traffic-group-1)
        will be used.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Service payload that you provide in
        the C(parameters) field.
    type: str
    version_added: 2.5
  metadata:
    description:
      - Metadata associated with the iApp service.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Service payload that you provide in
        the C(parameters) field.
    type: list
    version_added: 2.7
  description:
    description:
      - Description of the iApp service.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Service payload that you provide in
        the C(parameters) field.
    type: str
    version_added: 2.7
  device_group:
    description:
      - The device group for the iApp service.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Service payload that you provide in
        the C(parameters) field.
    type: str
    version_added: 2.7
extends_documentation_fragment: f5
author:
  - Tim Rupp (@caphrim007)
  - Wojciech Wypior (@wojtek0806)
'''

EXAMPLES = r'''
- name: Create HTTP iApp service from iApp template
  bigip_iapp_service:
    name: foo-service
    template: f5.http
    parameters: "{{ lookup('file', 'f5.http.parameters.json') }}"
    state: present
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  delegate_to: localhost

- name: Upgrade foo-service to v1.2.0rc4 of the f5.http template
  bigip_iapp_service:
    name: foo-service
    template: f5.http.v1.2.0rc4
    state: present
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  delegate_to: localhost

- name: Configure a service using parameters in YAML
  bigip_iapp_service:
    name: tests
    template: web_frontends
    state: present
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  delegate_to: localhost

- name: Re-configure a service whose underlying iApp was updated in place
  bigip_iapp_service:
    name: tests
    template: web_frontends
    force: yes
    state: present
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  delegate_to: localhost

- name: Try to remove the iApp template before the associated Service is removed
  bigip_iapp_template:
    name: web_frontends
    state: absent
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  register: result
  failed_when:
    - result is not success
    - "'referenced by one or more applications' not in result.msg"

- name: Configure a service using more complicated parameters
  bigip_iapp_service:
    name: tests
    template: web_frontends
    state: present
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
      lists:
        - name: irules__irules
          value:
            - foo
            - bar
      tables:
        - name: basic__snatpool_members
        - name: net__snatpool_members
        - name: optimizations__hosts
        - name: pool__hosts
          columnNames:
            - name
          rows:
            - row:
                - internal.company.bar
        - name: pool__members
          columnNames:
            - addr
            - port
            - connection_limit
          rows:
            - row:
                - "none"
                - 80
                - 0
        - name: server_pools__servers
  delegate_to: localhost

- name: Override metadata that may or may not exist in parameters
  bigip_iapp_service:
    name: foo-service
    template: f5.http
    parameters: "{{ lookup('file', 'f5.http.parameters.json') }}"
    metadata:
      - persist: yes
        name: data 1
      - persist: yes
        name: data 2
    state: present
    provider:
      user: admin
      password: secret
      server: lb.mydomain.com
  delegate_to: localhost
'''

RETURN = r'''
# only common fields returned
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.basic import env_fallback
from ansible.module_utils.six import iteritems

try:
    from library.module_utils.network.f5.bigip import F5RestClient
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import AnsibleF5Parameters
    from library.module_utils.network.f5.common import fq_name
    from library.module_utils.network.f5.common import f5_argument_spec
    from library.module_utils.network.f5.common import transform_name
    from library.module_utils.network.f5.common import flatten_boolean
    from library.module_utils.network.f5.urls import build_service_uri
except ImportError:
    from ansible.module_utils.network.f5.bigip import F5RestClient
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import AnsibleF5Parameters
    from ansible.module_utils.network.f5.common import fq_name
    from ansible.module_utils.network.f5.common import f5_argument_spec
    from ansible.module_utils.network.f5.common import transform_name
    from ansible.module_utils.network.f5.common import flatten_boolean
    from ansible.module_utils.network.f5.urls import build_service_uri


class Parameters(AnsibleF5Parameters):
    api_map = {
        'strictUpdates': 'strict_updates',
        'trafficGroup': 'traffic_group',
        'deviceGroup': 'device_group',
    }

    returnables = [
        'tables',
        'variables',
        'lists',
        'strict_updates',
        'traffic_group',
        'device_group',
        'metadata',
        'template',
        'description',
    ]

    api_attributes = [
        'tables',
        'variables',
        'template',
        'lists',
        'deviceGroup',
        'inheritedDevicegroup',
        'inheritedTrafficGroup',
        'trafficGroup',
        'strictUpdates',
        # 'metadata',
        'description',
    ]

    updatables = [
        'tables',
        'variables',
        'lists',
        'strict_updates',
        'device_group',
        'traffic_group',
        'metadata',
        'description',
    ]

    def to_return(self):
        result = {}
        for returnable in self.returnables:
            result[returnable] = getattr(self, returnable)
        result = self._filter_params(result)
        return result

    def normalize_tables(self, tables):
        result = []
        for table in tables:
            tmp = dict()
            name = table.get('name', None)
            if name is None:
                raise F5ModuleError(
                    "One of the provided tables does not have a name"
                )
            tmp['name'] = str(name)
            columns = table.get('columnNames', None)
            if columns:
                tmp['columnNames'] = [str(x) for x in columns]
                # You cannot have rows without columns
                rows = table.get('rows', None)
                if rows:
                    tmp['rows'] = []
                    for row in rows:
                        tmp['rows'].append(dict(row=[str(x) for x in row['row']]))
            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    def normalize_variables(self, variables):
        result = []
        for variable in variables:
            tmp = dict((str(k), str(v)) for k, v in iteritems(variable))
            if 'encrypted' not in tmp:
                # BIG-IP will inject an 'encrypted' key if you don't provide one.
                # If you don't provide one, then we give you the default 'no', by
                # default.
                tmp['encrypted'] = 'no'
            if 'value' not in tmp:
                tmp['value'] = ''

            # This seems to happen only on 12.0.0
            elif tmp['value'] == 'none':
                tmp['value'] = ''
            elif tmp['value'] == 'True':
                tmp['value'] = 'yes'
            elif tmp['value'] == 'False':
                tmp['value'] = 'no'
            elif isinstance(tmp['value'], bool):
                if tmp['value'] is True:
                    tmp['value'] = 'yes'
                else:
                    tmp['value'] = 'no'

            if tmp['encrypted'] == 'True':
                tmp['encrypted'] = 'yes'
            elif tmp['encrypted'] == 'False':
                tmp['encrypted'] = 'no'
            elif isinstance(tmp['encrypted'], bool):
                if tmp['encrypted'] is True:
                    tmp['encrypted'] = 'yes'
                else:
                    tmp['encrypted'] = 'no'

            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    def normalize_list(self, lists):
        result = []
        for list in lists:
            tmp = dict((str(k), str(v)) for k, v in iteritems(list) if k != 'value')
            if 'encrypted' not in list:
                # BIG-IP will inject an 'encrypted' key if you don't provide one.
                # If you don't provide one, then we give you the default 'no', by
                # default.
                tmp['encrypted'] = 'no'
            if 'value' in list:
                if len(list['value']) > 0:
                    # BIG-IP removes empty values entries, so mimic this behavior
                    # for user-supplied values.
                    tmp['value'] = [str(x) for x in list['value']]

            if tmp['encrypted'] == 'True':
                tmp['encrypted'] = 'yes'
            elif tmp['encrypted'] == 'False':
                tmp['encrypted'] = 'no'
            elif isinstance(tmp['encrypted'], bool):
                if tmp['encrypted'] is True:
                    tmp['encrypted'] = 'yes'
                else:
                    tmp['encrypted'] = 'no'

            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    def normalize_metadata(self, metadata):
        result = []
        for item in metadata:
            name = item.get('name', None)
            persist = flatten_boolean(item.get('persist', "no"))
            if persist == "yes":
                persist = "true"
            else:
                persist = "false"
            result.append({
                "name": name,
                "persist": persist
            })
        return result


class ApiParameters(Parameters):
    @property
    def metadata(self):
        if self._values['metadata'] is None:
            return None
        return self._values['metadata']

    @property
    def tables(self):
        if self._values['tables'] is None:
            return None
        return self.normalize_tables(self._values['tables'])

    @property
    def lists(self):
        if self._values['lists'] is None:
            return None
        return self.normalize_list(self._values['lists'])

    @property
    def variables(self):
        if self._values['variables'] is None:
            return None
        return self.normalize_variables(self._values['variables'])

    @property
    def device_group(self):
        if self._values['device_group'] in [None, 'none']:
            return None
        return self._values['device_group']


class ModuleParameters(Parameters):
    @property
    def param_lists(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('lists', None)
        return result

    @property
    def param_tables(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('tables', None)
        return result

    @property
    def param_variables(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('variables', None)
        return result

    @property
    def param_metadata(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('metadata', None)
        return result

    @property
    def param_description(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('description', None)
        return result

    @property
    def param_traffic_group(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('trafficGroup', None)
        if not result:
            return result
        return fq_name(self.partition, result)

    @property
    def param_device_group(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('deviceGroup', None)
        if not result:
            return result
        return fq_name(self.partition, result)

    @property
    def param_strict_updates(self):
        if self._values['parameters'] is None:
            return None
        result = self._values['parameters'].get('strictUpdates', None)
        return flatten_boolean(result)

    @property
    def tables(self):
        if self._values['tables']:
            return self.normalize_tables(self._values['tables'])
        elif self.param_tables:
            return self.normalize_tables(self.param_tables)
        return None

    @property
    def lists(self):
        if self._values['lists']:
            return self.normalize_list(self._values['lists'])
        elif self.param_lists:
            return self.normalize_list(self.param_lists)
        return None

    @property
    def variables(self):
        if self._values['variables']:
            return self.normalize_variables(self._values['variables'])
        elif self.param_variables:
            return self.normalize_variables(self.param_variables)
        return None

    @property
    def metadata(self):
        if self._values['metadata']:
            result = self.normalize_metadata(self._values['metadata'])
        elif self.param_metadata:
            result = self.normalize_metadata(self.param_metadata)
        else:
            return None
        return result

    @property
    def template(self):
        if self._values['template'] is None:
            return None
        return fq_name(self.partition, self._values['template'])

    @property
    def device_group(self):
        if self._values['device_group'] not in [None, 'none']:
            result = fq_name(self.partition, self._values['device_group'])
        elif self.param_device_group not in [None, 'none']:
            result = self.param_device_group
        else:
            return None
        if not result.startswith('/Common/'):
            raise F5ModuleError(
                "Device groups can only exist in /Common"
            )
        return result

    @property
    def traffic_group(self):
        if self._values['traffic_group']:
            result = fq_name(self.partition, self._values['traffic_group'])
        elif self.param_traffic_group:
            result = self.param_traffic_group
        else:
            return None
        if not result.startswith('/Common/'):
            raise F5ModuleError(
                "Traffic groups can only exist in /Common"
            )
        return result

    @property
    def strict_updates(self):
        if self._values['strict_updates'] is not None:
            result = flatten_boolean(self._values['strict_updates'])
        elif self.param_strict_updates is not None:
            result = flatten_boolean(self.param_strict_updates)
        else:
            return None
        if result == 'yes':
            return 'enabled'
        return 'disabled'

    @property
    def description(self):
        if self._values['description']:
            return self._values['description']
        elif self.param_description:
            return self.param_description
        return None


class Changes(Parameters):
    pass


class UsableChanges(Changes):
    pass


class ReportableChanges(Changes):
    pass


class Difference(object):
    def __init__(self, want, have=None):
        self.want = want
        self.have = have

    def compare(self, param):
        try:
            result = getattr(self, param)
            return result
        except AttributeError:
            return self.__default(param)

    def __default(self, param):
        attr1 = getattr(self.want, param)
        try:
            attr2 = getattr(self.have, param)
            if attr1 != attr2:
                return attr1
        except AttributeError:
            return attr1

    @property
    def metadata(self):
        if self.want.metadata is None:
            return None
        if self.have.metadata is None:
            return self.want.metadata
        want = [(k, v) for d in self.want.metadata for k, v in iteritems(d)]
        have = [(k, v) for d in self.have.metadata for k, v in iteritems(d)]
        if set(want) != set(have):
            return dict(
                metadata=self.want.metadata
            )


class ModuleManager(object):
    def __init__(self, *args, **kwargs):
        self.module = kwargs.get('module', None)
        self.client = F5RestClient(**self.module.params)
        self.have = None
        self.want = ModuleParameters(params=self.module.params)
        self.changes = UsableChanges()

    def _set_changed_options(self):
        changed = {}
        for key in Parameters.returnables:
            if getattr(self.want, key) is not None:
                changed[key] = getattr(self.want, key)
        if changed:
            self.changes = UsableChanges(params=changed)

    def _update_changed_options(self):
        diff = Difference(self.want, self.have)
        updatables = Parameters.updatables
        changed = dict()
        for k in updatables:
            change = diff.compare(k)
            if change is None:
                continue
            else:
                if isinstance(change, dict):
                    changed.update(change)
                else:
                    changed[k] = change
        if changed:
            self.changes = UsableChanges(params=changed)
            return True
        return False

    def exec_module(self):
        changed = False
        result = dict()
        state = self.want.state

        if state == "present":
            changed = self.present()
        elif state == "absent":
            changed = self.absent()

        changes = self.changes.to_return()
        result.update(**changes)
        result.update(dict(changed=changed))
        return result

    def present(self):
        if self.exists():
            return self.update()
        else:
            return self.create()

    def absent(self):
        if self.exists():
            return self.remove()
        return False

    def create(self):
        self._set_changed_options()
        if self.want.traffic_group is None:
            self.want.update({'traffic_group': '/Common/traffic-group-1'})
        if not self.template_exists():
            raise F5ModuleError(
                "The specified template does not exist in the provided partition."
            )
        if self.module.check_mode:
            return True
        self.create_on_device()
        return True

    def remove(self):
        if self.module.check_mode:
            return True
        self.remove_from_device()
        if self.exists():
            raise F5ModuleError("Failed to delete the iApp service")
        return True

    def update(self):
        self.have = self.read_current_from_device()
        if not self.should_update() and not self.want.force:
            return False
        if self.module.check_mode:
            return True
        self.update_on_device()
        return True

    def should_update(self):
        result = self._update_changed_options()
        if result:
            return True
        return False

    def exists(self):
        base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
            self.client.provider['server'],
            self.client.provider['server_port']
        )
        uri = build_service_uri(base_uri, self.want.partition, self.want.name)
        resp = self.client.api.get(uri)
        try:
            response = resp.json()
        except ValueError:
            return False
        if resp.status == 404 or 'code' in response and response['code'] == 404:
            return False
        return True

    def update_on_device(self):
        params = self.changes.api_params()
        if params:
            params['execute-action'] = 'definition'
            base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
                self.client.provider['server'],
                self.client.provider['server_port']
            )
            uri = build_service_uri(base_uri, self.want.partition, self.want.name)
            resp = self.client.api.patch(uri, json=params)
            try:
                response = resp.json()
            except ValueError as ex:
                raise F5ModuleError(str(ex))

            if 'code' in response and response['code'] == 400:
                if 'message' in response:
                    raise F5ModuleError(response['message'])
                else:
                    raise F5ModuleError(resp.content)

        if self.changes.metadata:
            params = dict(metadata=self.changes.metadata)
            params.update({'execute-action': 'definition'})
            base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
                self.client.provider['server'],
                self.client.provider['server_port']
            )
            uri = build_service_uri(base_uri, self.want.partition, self.want.name)
            resp = self.client.api.patch(uri, json=params)

            try:
                response = resp.json()
            except ValueError as ex:
                raise F5ModuleError(str(ex))

            if 'code' in response and response['code'] == 400:
                if 'message' in response:
                    raise F5ModuleError(response['message'])
                else:
                    raise F5ModuleError(resp.content)

    def read_current_from_device(self):
        base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
            self.client.provider['server'],
            self.client.provider['server_port']
        )
        uri = build_service_uri(base_uri, self.want.partition, self.want.name)
        resp = self.client.api.get(uri)

        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)
        return ApiParameters(params=response)

    def template_exists(self):
        name = fq_name(self.want.partition, self.want.template)
        parts = name.split('/')
        uri = "https://{0}:{1}/mgmt/tm/sys/application/template/{2}".format(
            self.client.provider['server'],
            self.client.provider['server_port'],
            transform_name(parts[1], parts[2])
        )
        resp = self.client.api.get(uri)
        try:
            response = resp.json()
        except ValueError:
            return False
        if resp.status == 404 or 'code' in response and response['code'] == 404:
            return False
        return True

    def create_on_device(self):
        params = self.changes.api_params()
        params['name'] = self.want.name
        params['partition'] = self.want.partition
        uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
            self.client.provider['server'],
            self.client.provider['server_port']
        )
        resp = self.client.api.post(uri, json=params)

        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] in [400, 403]:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)

        if self.changes.metadata:
            payload = dict(metadata=self.changes.metadata)
            base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
                self.client.provider['server'],
                self.client.provider['server_port']
            )
            uri = build_service_uri(base_uri, self.want.partition, self.want.name)
            resp = self.client.api.patch(uri, json=payload)
            try:
                response = resp.json()
            except ValueError as ex:
                raise F5ModuleError(str(ex))

            if 'code' in response and response['code'] == 400:
                if 'message' in response:
                    raise F5ModuleError(response['message'])
                else:
                    raise F5ModuleError(resp.content)

    def remove_from_device(self):
        base_uri = "https://{0}:{1}/mgmt/tm/sys/application/service/".format(
            self.client.provider['server'],
            self.client.provider['server_port']
        )
        uri = build_service_uri(base_uri, self.want.partition, self.want.name)

        # Metadata needs to be zero'd before the service is removed because
        # otherwise, the API will error out saying that "configuration items"
        # currently exist.
        #
        # In other words, the REST API is not able to delete a service while
        # there is existing metadata
        payload = dict(metadata=[])
        resp = self.client.api.patch(uri, json=payload)
        try:
            response = resp.json()
        except ValueError as ex:
            raise F5ModuleError(str(ex))

        if 'code' in response and response['code'] == 400:
            if 'message' in response:
                raise F5ModuleError(response['message'])
            else:
                raise F5ModuleError(resp.content)

        resp = self.client.api.delete(uri)

        if resp.status == 200:
            return True
        raise F5ModuleError(resp.content)


class ArgumentSpec(object):
    def __init__(self):
        self.supports_check_mode = True
        argument_spec = dict(
            name=dict(required=True),
            template=dict(),
            description=dict(),
            device_group=dict(),
            parameters=dict(
                type='dict'
            ),
            state=dict(
                default='present',
                choices=['absent', 'present']
            ),
            force=dict(
                default='no',
                type='bool'
            ),
            strict_updates=dict(
                type='bool',
                default='yes'
            ),
            metadata=dict(type='list'),
            traffic_group=dict(),
            partition=dict(
                default='Common',
                fallback=(env_fallback, ['F5_PARTITION'])
            )
        )
        self.argument_spec = {}
        self.argument_spec.update(f5_argument_spec)
        self.argument_spec.update(argument_spec)


def main():
    spec = ArgumentSpec()

    module = AnsibleModule(
        argument_spec=spec.argument_spec,
        supports_check_mode=spec.supports_check_mode
    )

    try:
        mm = ModuleManager(module=module)
        results = mm.exec_module()
        module.exit_json(**results)
    except F5ModuleError as ex:
        module.fail_json(msg=str(ex))


if __name__ == '__main__':
    main()



# Copyright (c) 2018 NetApp
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

''' unit tests for module_utils netapp.py '''
from __future__ import absolute_import, division, print_function

import pytest

from ansible.module_utils.ansible_release import __version__ as ansible_version

import ansible.module_utils.netapp as netapp_utils

if not netapp_utils.has_netapp_lib():
    pytestmark = pytest.mark.skip("skipping as missing required netapp_lib")


class MockONTAPConnection(object):
    ''' mock a server connection to ONTAP host '''

    def __init__(self, kind=None, parm1=None):
        ''' save arguments '''
        self.type = kind
        self.parm1 = parm1
        self.xml_in = None
        self.xml_out = None

    def invoke_successfully(self, xml, enable_tunneling):  # pylint: disable=unused-argument
        ''' mock invoke_successfully returning xml data '''
        self.xml_in = xml
        if self.type == 'vserver':
            xml = self.build_vserver_info(self.parm1)
        self.xml_out = xml
        return xml

    @staticmethod
    def build_vserver_info(vserver):
        ''' build xml data for vserser-info '''
        xml = netapp_utils.zapi.NaElement('xml')
        attributes = netapp_utils.zapi.NaElement('attributes-list')
        attributes.add_node_with_children('vserver-info',
                                          **{'vserver-name': vserver})
        xml.add_child_elem(attributes)
        return xml


def test_ems_log_event_version():
    ''' validate Ansible version is correctly read '''
    source = 'unittest'
    server = MockONTAPConnection()
    netapp_utils.ems_log_event(source, server)
    xml = server.xml_in
    version = xml.get_child_content('app-version')
    assert version == ansible_version
    print("Ansible version: %s" % ansible_version)


def test_get_cserver():
    ''' validate cluster vserser name is correctly retrieved '''
    svm_name = 'svm1'
    server = MockONTAPConnection('vserver', svm_name)
    cserver = netapp_utils.get_cserver(server)
    assert cserver == svm_name



# Copyright 2018 Fortinet, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <https://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import json
from ansible.module_utils.network.fortimanager.fortimanager import FortiManagerHandler
import pytest

try:
    from ansible.modules.network.fortimanager import fmgr_fwobj_vip
except ImportError:
    pytest.skip("Could not load required modules for testing", allow_module_level=True)


def load_fixtures():
    fixture_path = os.path.join(os.path.dirname(__file__), 'fixtures') + "/{filename}.json".format(
        filename=os.path.splitext(os.path.basename(__file__))[0])
    try:
        with open(fixture_path, "r") as fixture_file:
            fixture_data = json.load(fixture_file)
    except IOError:
        return []
    return [fixture_data]


@pytest.fixture(autouse=True)
def module_mock(mocker):
    connection_class_mock = mocker.patch('ansible.module_utils.basic.AnsibleModule')
    return connection_class_mock


@pytest.fixture(autouse=True)
def connection_mock(mocker):
    connection_class_mock = mocker.patch('ansible.modules.network.fortimanager.fmgr_fwobj_vip.Connection')
    return connection_class_mock


@pytest.fixture(scope="function", params=load_fixtures())
def fixture_data(request):
    func_name = request.function.__name__.replace("test_", "")
    return request.param.get(func_name, None)


fmg_instance = FortiManagerHandler(connection_mock, module_mock)


def test_fmgr_firewall_vip_modify(fixture_data, mocker):
    mocker.patch("ansible.module_utils.network.fortimanager.fortimanager.FortiManagerHandler.process_request",
                 side_effect=fixture_data)
    #  Fixture sets used:###########################

    ##################################################
    # comment: Created by Ansible
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # mapped-addr: None
    # ssl-client-session-state-timeout: None
    # src-filter: None
    # server-type: None
    # ssl-hpkp-include-subdomains: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: 10.7.220.25
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: tcp
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # ssl-client-renegotiation: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # weblogic-server: None
    # http-cookie-share: None
    # color: 17
    # ssl-mode: None
    # portforward: enable
    # http-multiplex: None
    # http-cookie-generation: None
    # ssl-client-fallback: None
    # extip: 82.72.192.185
    # extintf: any
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None

    # adom: ansible
    # ssl-client-session-state-max: None
    # http-ip-header: None
    # http-ip-header-name: None
    # ssl-certificate: None
    # ssl-hsts: None
    # arp-reply: None
    # ssl-hsts-include-subdomains: None
    # ssl-min-version: None
    # ldb-method: None
    # ssl-server-session-state-timeout: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: 443
    # name: Basic PNAT Map Port 10443
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # outlook-web-access: None
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-server-max-version: None
    # ssl-hpkp-report-uri: None
    # http-cookie-domain-from-host: None
    # ssl-algorithm: None
    # gratuitous-arp-interval: None
    # extport: 10443
    # max-embryonic-connections: None
    # mode: set
    # http-cookie-path: None
    # ssl-pfs: None
    # ssl-server-algorithm: None
    ##################################################
    ##################################################
    # comment: Created by Ansible
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # ssl-hsts-include-subdomains: None
    # mapped-addr: None
    # src-filter: None
    # server-type: None
    # mode: set
    # ssl-hpkp-include-subdomains: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: 3.3.3.0/24, 4.0.0.0/24
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # ssl-client-renegotiation: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # adom: ansible
    # http-cookie-share: None
    # ssl-server-session-state-timeout: None
    # color: 12
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # max-embryonic-connections: None
    # ssl-client-fallback: None
    # ssl-hpkp-report-uri: None
    # extip: 192.168.0.1-192.168.0.100
    # extintf: dmz
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None
    # http-ip-header-name: None
    # weblogic-server: None
    # ssl-client-session-state-max: None
    # http-ip-header: None

    # ssl-hsts: None
    # arp-reply: None
    # extaddr: None
    # ssl-min-version: None
    # ldb-method: None
    # ssl-certificate: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # outlook-web-access: None
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # name: Basic DNS Translation
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-pfs: None
    # ssl-server-max-version: None
    # ssl-client-session-state-timeout: None
    # http-cookie-domain-from-host: None
    # extport: None
    # ssl-server-algorithm: None
    # gratuitous-arp-interval: None
    # http-cookie-path: None
    # ssl-algorithm: None
    # http-multiplex: None
    ##################################################
    ##################################################
    # comment: Created by Ansible
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # mapped-addr: google-play
    # ssl-client-session-state-timeout: None
    # src-filter: None
    # ldb-method: None
    # server-type: None
    # ssl-hpkp-include-subdomains: None
    # ssl-client-renegotiation: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: None
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # weblogic-server: None
    # http-cookie-share: None
    # color: 5
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # ssl-client-fallback: None
    # extip: None
    # extintf: None
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None

    # adom: ansible
    # ssl-client-session-state-max: None
    # http-ip-header: None
    # http-ip-header-name: None
    # ssl-certificate: None
    # ssl-hsts: None
    # arp-reply: None
    # extport: None
    # ssl-min-version: None
    # ssl-server-algorithm: None
    # ssl-server-session-state-timeout: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # name: Basic FQDN Translation
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # outlook-web-access: None
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-server-max-version: None
    # ssl-hpkp-report-uri: None
    # http-cookie-domain-from-host: None
    # ssl-algorithm: None
    # gratuitous-arp-interval: None
    # ssl-hsts-include-subdomains: None
    # max-embryonic-connections: None
    # mode: set
    # http-cookie-path: None
    # ssl-pfs: None
    # http-multiplex: None
    ##################################################
    ##################################################
    # comment: Created by Ansible
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # mapped-addr: None
    # src-filter: None
    # server-type: None
    # mode: set
    # ssl-hpkp-include-subdomains: None
    # extport: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: 10.7.220.25
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # ssl-server-algorithm: None
    # extaddr: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # adom: ansible
    # http-cookie-share: None
    # ssl-server-session-state-timeout: None
    # color: 17
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # max-embryonic-connections: None
    # ssl-client-fallback: None
    # ssl-hpkp-report-uri: None
    # extip: 82.72.192.185
    # extintf: any
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None
    # http-ip-header-name: None
    # weblogic-server: None
    # ssl-client-session-state-max: None
    # http-ip-header: None

    # ssl-hsts: None
    # arp-reply: None
    # ssl-client-renegotiation: None
    # ssl-min-version: None
    # ldb-method: None
    # ssl-certificate: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # outlook-web-access: None
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # name: Basic StaticNAT Map
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-pfs: None
    # ssl-client-session-state-timeout: None
    # http-cookie-domain-from-host: None
    # ssl-hsts-include-subdomains: None
    # ssl-server-max-version: None
    # gratuitous-arp-interval: None
    # http-cookie-path: None
    # ssl-algorithm: None
    # http-multiplex: None
    ##################################################
    ##################################################
    # comment: Created by Ansible
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # mapped-addr: None
    # ssl-client-session-state-timeout: None
    # src-filter: None
    # server-type: None
    # ssl-hpkp-include-subdomains: None
    # ssl-client-renegotiation: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: 10.7.220.25
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: tcp
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # weblogic-server: None
    # http-cookie-share: None
    # color: 17
    # ssl-mode: None
    # portforward: enable
    # http-cookie-generation: None
    # ssl-client-fallback: None
    # extip: 82.72.192.185
    # extintf: any
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None

    # adom: ansible
    # ssl-client-session-state-max: None
    # http-ip-header: None
    # http-ip-header-name: None
    # ssl-min-version: None
    # ssl-certificate: None
    # ssl-hsts: None
    # arp-reply: None
    # ssl-hsts-include-subdomains: None
    # http-multiplex: None
    # ldb-method: None
    # ssl-server-session-state-timeout: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: 443
    # name: Basic PNAT Map Port 10443
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # outlook-web-access: None
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-server-max-version: None
    # ssl-hpkp-report-uri: None
    # http-cookie-domain-from-host: None
    # ssl-algorithm: None
    # gratuitous-arp-interval: None
    # extport: 10443
    # max-embryonic-connections: None
    # mode: set
    # http-cookie-path: None
    # ssl-pfs: None
    # ssl-server-algorithm: None
    ##################################################
    ##################################################
    # comment: None
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # ssl-hpkp: None
    # ssl-hsts-include-subdomains: None
    # mapped-addr: None
    # src-filter: None
    # server-type: None
    # mode: delete
    # ssl-hpkp-include-subdomains: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # mappedip: None
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # adom: ansible
    # http-cookie-share: None
    # ssl-server-session-state-timeout: None
    # color: None
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # max-embryonic-connections: None
    # ssl-client-fallback: None
    # ssl-hpkp-report-uri: None
    # extip: None
    # extintf: None
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None
    # http-ip-header-name: None
    # weblogic-server: None
    # ssl-client-session-state-max: None
    # http-ip-header: None

    # ssl-hsts: None
    # arp-reply: None
    # ssl-client-renegotiation: None
    # http-multiplex: None
    # ldb-method: None
    # ssl-certificate: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # outlook-web-access: None
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # name: Basic PNAT Map Port 10443
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-pfs: None
    # ssl-server-max-version: None
    # ssl-client-session-state-timeout: None
    # http-cookie-domain-from-host: None
    # extport: None
    # ssl-server-algorithm: None
    # gratuitous-arp-interval: None
    # http-cookie-path: None
    # ssl-algorithm: None
    # ssl-min-version: None
    ##################################################
    ##################################################
    # comment: None
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # mappedip: None
    # mapped-addr: None
    # ssl-client-session-state-timeout: None
    # src-filter: None
    # ldb-method: None
    # server-type: None
    # ssl-hpkp-include-subdomains: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # ssl-hpkp: None
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # ssl-client-renegotiation: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # weblogic-server: None
    # http-cookie-share: None
    # color: None
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # ssl-client-fallback: None
    # extip: None
    # extintf: None
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None

    # adom: ansible
    # ssl-client-session-state-max: None
    # http-ip-header: None
    # http-ip-header-name: None
    # ssl-certificate: None
    # ssl-hsts: None
    # arp-reply: None
    # extport: None
    # http-multiplex: None
    # ssl-server-algorithm: None
    # ssl-server-session-state-timeout: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # name: Basic StaticNAT Map
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-hpkp-primary: None
    # outlook-web-access: None
    # ssl-server-session-state-type: None
    # ssl-client-session-state-type: None

    # ssl-http-match-host: None

    # ssl-server-max-version: None
    # ssl-hpkp-report-uri: None
    # http-cookie-domain-from-host: None
    # ssl-algorithm: None
    # gratuitous-arp-interval: None
    # ssl-hsts-include-subdomains: None
    # max-embryonic-connections: None
    # mode: delete
    # http-cookie-path: None
    # ssl-pfs: None
    # ssl-min-version: None
    ##################################################
    ##################################################
    # comment: None
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # mappedip: None
    # mapped-addr: None
    # src-filter: None
    # server-type: None
    # mode: delete
    # ssl-hpkp-include-subdomains: None
    # extport: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # ssl-hpkp: None
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # ssl-server-algorithm: None
    # ssl-client-renegotiation: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # adom: ansible
    # http-cookie-share: None
    # ssl-server-session-state-timeout: None
    # color: None
    # ssl-mode: None
    # portforward: None
    # http-multiplex: None
    # http-cookie-generation: None
    # max-embryonic-connections: None
    # ssl-client-fallback: None
    # ssl-hpkp-report-uri: None
    # extip: None
    # extintf: None
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None
    # http-ip-header-name: None
    # weblogic-server: None
    # ssl-client-session-state-max: None
    # http-ip-header: None
    # ssl-hsts: None
    # arp-reply: None
    # extaddr: None
    # ssl-hpkp-primary: None
    # ldb-method: None
    # ssl-certificate: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # outlook-web-access: None
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-client-session-state-type: None
    # name: Basic DNS Translation
    # ssl-server-session-state-type: None

    # ssl-http-match-host: None
    # ssl-pfs: None
    # ssl-client-session-state-timeout: None
    # http-cookie-domain-from-host: None
    # ssl-hsts-include-subdomains: None
    # ssl-server-max-version: None
    # gratuitous-arp-interval: None
    # http-cookie-path: None
    # ssl-algorithm: None
    # ssl-min-version: None
    ##################################################
    ##################################################
    # ldb-method: None
    # ssl-send-empty-frags: None
    # srcintf-filter: None
    # ssl-max-version: None
    # ssl-server-session-state-max: None
    # mappedip: None
    # ssl-hsts: None
    # mapped-addr: None
    # src-filter: None
    # server-type: None
    # ssl-hpkp-include-subdomains: None
    # ssl-client-renegotiation: None
    # ssl-http-location-conversion: None
    # https-cookie-secure: None
    # extip: None
    # ssl-hpkp: None
    # ssl-server-cipher-suites: {'priority': None, 'cipher': None, 'versions': None}
    # protocol: None
    # ssl-hpkp-backup: None
    # ssl-dh-bits: None
    # dns-mapping-ttl: None
    # ssl-hsts-age: None
    # extaddr: None
    # ssl-hpkp-primary: None
    # monitor: None
    # service: None
    # ssl-hpkp-age: None
    # http-cookie-age: None
    # weblogic-server: None
    # http-cookie-share: None
    # name: Basic FQDN Translation
    # color: None
    # ssl-mode: None
    # portforward: None
    # http-cookie-generation: None
    # ssl-client-fallback: None

    # http-ip-header: None
    # persistence: None
    # websphere-server: None
    # nat-source-vip: None
    # portmapping-type: None
    # adom: ansible
    # ssl-client-session-state-max: None
    # extintf: None
    # ssl-server-max-version: None
    # http-ip-header-name: None
    # ssl-certificate: None
    # ssl-server-session-state-type: None
    # arp-reply: None
    # ssl-hsts-include-subdomains: None
    # ssl-min-version: None
    # ssl-server-algorithm: None
    # ssl-server-session-state-timeout: None
    # ssl-server-min-version: None
    # http-cookie-domain: None
    # mappedport: None
    # outlook-web-access: None
    # ssl-cipher-suites: {'cipher': None, 'versions': None}
    # ssl-client-session-state-type: None
    # ssl-http-match-host: None

    # ssl-client-session-state-timeout: None
    # comment: None
    # ssl-hpkp-report-uri: None
    # http-cookie-domain-from-host: None
    # ssl-algorithm: None
    # gratuitous-arp-interval: None
    # extport: None
    # max-embryonic-connections: None
    # mode: delete
    # http-cookie-path: None
    # ssl-pfs: None
    # http-multiplex: None
    ##################################################

    # Test using fixture 1 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[0]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 2 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[1]['paramgram_used'])
    assert output['raw_response']['status']['code'] == -10131
    # Test using fixture 3 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[2]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 4 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[3]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 5 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[4]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 6 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[5]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 7 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[6]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
    # Test using fixture 8 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[7]['paramgram_used'])
    assert output['raw_response']['status']['code'] == -3
    # Test using fixture 9 #
    output = fmgr_fwobj_vip.fmgr_firewall_vip_modify(fmg_instance, fixture_data[8]['paramgram_used'])
    assert output['raw_response']['status']['code'] == 0
